import os
from pathlib import Path

from luna.util.exceptions import IllegalArgumentError
from luna.util.file import (is_file_valid, get_file_format,
                            create_directory, new_unique_filename,
                            remove_directory)
from luna.wrappers.obabel import convert_molecule
from luna.wrappers.base import MolWrapper
from luna.wrappers.chimera import add_charges

import logging
logger = logging.getLogger()


class PrepareLigand:

    def __init__(self, add_h=False, ph=7.4,
                 use_canonical_smiles=False,
                 add_charges=False, charge_method="am1-bcc",
                 minimize=False, steps=150, forcefield="MMFF94",
                 use_sd=True, min_opts=None):

        self.add_h = add_h
        self.ph = ph

        # If Smiles are provided, transform it to their canonical form.
        self.use_canonical_smiles = use_canonical_smiles

        # Arguments to add charges with Antechamber.
        self.add_charges = add_charges
        self.charge_method = charge_method

        # Arguments to minimize ligand with Open Babel.
        self.minimize = minimize
        self.steps = steps
        self.forcefield = forcefield
        self.use_sd = use_sd
        self.min_opts = min_opts or {}

    def prep(self, ligand_input, output_file, is_smiles=False,
             ligand_name=None):

        if not is_smiles:
            ligand_input = str(Path(ligand_input).absolute())

        output_path = Path(output_file).parent.absolute()
        filename = Path(output_file).stem

        cur_path = os.path.abspath(os.getcwd())

        tmp_path = new_unique_filename(output_path)
        create_directory(tmp_path)

        os.chdir(tmp_path)

        ligand_file = None
        try:
            if is_smiles:
                if not isinstance(ligand_input, str):
                    raise TypeError("Invalid type for the ligand SMILES. "
                                    "A string was expected instead.")

                # Transform Smiles to its canonical form if required.
                if self.use_canonical_smiles:
                    ligand_input = self.to_canonical_form(ligand_input)

                # Add hydrogen if required.
                if self.add_h:
                    ligand_input = self.add_h_to_ligand(ligand_input)

                ligand_file = f"{tmp_path}/{filename}_coords.mol2"
                self.generate_3d_coords(ligand_input, ligand_file)

            elif isinstance(ligand_input, str):
                if not is_file_valid(ligand_input):
                    raise IllegalArgumentError("'%s' does not exist or "
                                               "is not a file." % ligand_input)

                ligand_file = ligand_input
            else:
                raise IllegalArgumentError("'ligand_input' must be a "
                                           "string representing a molecular "
                                           "file or SMILES.")

            if self.add_charges:
                tmp_file = f"{tmp_path}/{filename}_charged.mol2"
                self.add_charges_to_ligand(ligand_file, tmp_file)
                ligand_file = tmp_file

            if self.minimize:
                tmp_file = f"{tmp_path}/{filename}_min.mol2"
                self.minimize_ligand(ligand_file, tmp_file)
                ligand_file = tmp_file

            # Molecular file format.
            ext = get_file_format(ligand_file, 1)

            if ligand_name:
                tmp_file = f"{tmp_path}/{filename}_renamed.{ext}"
                self.rename_ligand_at_file(ligand_name, ligand_file, tmp_file)
                ligand_file = tmp_file

            # Update the charge method in the resulting file.
            if ext == "mol2" and self.add_charges:
                self.set_charge_method(ligand_file)

        except Exception as e:
            ligand_file = None
            logger.exception(e)
            raise

        # Safe finish: set back the working directory to its initial value.
        finally:
            os.chdir(cur_path)
            if ligand_file and os.path.exists(ligand_file):
                os.rename(ligand_file, Path(output_file).absolute())
            remove_directory(tmp_path)

    def to_canonical_form(self, ligand_input):
        return convert_molecule(ligand_input, input_format="smi",
                                output_format="can")

    def add_h_to_ligand(self, ligand_input):
        # If it is a string, but not a file path.
        if not is_file_valid(ligand_input):
            if self.ph is None:
                opts = {"h": ""}
            else:
                opts = {"p": self.ph}

            return convert_molecule(ligand_input, input_format="smi",
                                    output_format="smi", opts=opts)
        else:
            # Not implemented yet.
            pass

    def generate_3d_coords(self, input_file, output_file=None):
        if not isinstance(input_file, str):
            raise IllegalArgumentError("This function accepts only "
                                       "SMILES string.")

        return convert_molecule(input_file, input_format="smi",
                                output_file=output_file, output_format="mol2",
                                opts={"gen3d": ""})

    def rename_ligand_at_file(self, new_name, input_file, output_file=None):
        output_file = output_file or input_file

        output_format = get_file_format(output_file, 1)
        target_row = 0
        if output_format == "mol2":
            target_row = 1

        lines = []
        with open(input_file, "r") as IN:
            c = 0
            for line in IN:
                if c == target_row:
                    lines.append(new_name + "\n")
                else:
                    lines.append(line)

                c += 1

        with open(output_file, "w") as OUT:
            OUT.write("".join(lines))

    def remove_unity_records(self, input_file, output_file=None):
        output_file = output_file or input_file

        lines = []
        with open(input_file, "r") as IN:
            ignore = False
            for line in IN:
                if line.strip() == "@<TRIPOS>UNITY_ATOM_ATTR":
                    ignore = True
                elif line.strip() == "@<TRIPOS>BOND":
                    ignore = False

                if not ignore:
                    lines.append(line)

        with open(output_file, "w") as OUT:
            OUT.write("".join(lines))

    def get_unity_records(self, input_file):
        lines = []
        with open(input_file, "r") as IN:
            ignore = True
            for line in IN:
                if line.strip() == "@<TRIPOS>UNITY_ATOM_ATTR":
                    ignore = False
                elif line.strip() == "@<TRIPOS>BOND":
                    break

                if not ignore:
                    lines.append(line.rstrip("\n"))
        return lines

    def add_charges_to_ligand(self, input_file, output_file):
        input_format = get_file_format(input_file, 1)

        try:
            mol_obj = MolWrapper.from_mol_file(input_file, input_format,
                                               mol_obj_type="rdkit")
            total_charge = mol_obj.get_total_charge()
        except Exception:
            mol_obj = MolWrapper.from_mol_file(input_file, input_format,
                                               mol_obj_type="openbabel")
            total_charge = mol_obj.get_total_charge()

        add_charges(input_file, output_file, total_charge, self.charge_method)

        if input_format == "mol2":
            # Add charge information (Unity records) to the output file
            # so that the molecule can be parsed by Open Babel without
            # raising kekulization issues.
            unity_records = self.get_unity_records(input_file)

            lines = []
            with open(output_file) as IN:
                for line in IN:
                    line = line.rstrip("\n")

                    if line == "@<TRIPOS>BOND":
                        lines.extend(unity_records)

                    lines.append(line)

            with open(output_file, "w") as OUT:
                OUT.write("\n".join(lines))

        self._compare_mols(input_file, output_file)

    def _compare_mols(self, mol_file1, mol_file2):
        mol_format1 = get_file_format(mol_file1, 1)
        mol_format2 = get_file_format(mol_file2, 1)

        # Molecule 1
        mol_obj1 = MolWrapper.from_mol_file(mol_file1, mol_format1,
                                            mol_obj_type="openbabel")
        #
        # Molecule 2
        mol_obj2 = MolWrapper.from_mol_file(mol_file2, mol_format2,
                                            mol_obj_type="openbabel")

        try:
            mol_obj1 = MolWrapper(mol_obj1.as_rdkit())
            mol_obj2 = MolWrapper(mol_obj2.as_rdkit())

            #
            # Maximum Common Substructure.
            mcs = mol_obj1.find_mcs(mol_obj2)

            assert (mcs.numAtoms == mol_obj1.get_num_heavy_atoms()
                    and mcs.numBonds == len(mol_obj1.get_bonds()))

        except AssertionError:
            raise AssertionError("Molecules in the files '%s' and '%s' does "
                                 "not have a structural match."
                                 % (mol_file1, mol_file2))
        except Exception:
            # Compare molecules by InChI identifiers.
            inchi1 = convert_molecule(mol_file1, input_format="mol2",
                                      output_format="inchi")
            inchi2 = convert_molecule(mol_file2, input_format="mol2",
                                      output_format="inchi")

            if inchi1 != inchi2:
                raise AssertionError("Molecules in the files '%s' and '%s' "
                                     "does not have equal InChI identifiers."
                                     % (mol_file1, mol_file2))

    def minimize_ligand(self, input_file, output_file=None):
        opts = {"minimize": "",
                "steps": self.steps,
                "ff": self.forcefield}

        if self.use_sd:
            opts["sd"] = ""

        input_format = get_file_format(input_file, 1)

        return convert_molecule(input_file, input_format=input_format,
                                output_file=output_file, output_format="mol2",
                                opts=opts)

    def set_charge_method(self, input_file, output_file=None):
        output_file = output_file or input_file

        lines = []
        with open(input_file) as IN:
            cm = "GASTEIGER"
            if self.charge_method == "am1-bcc":
                cm = "AM1-BCC"

            c = 0
            for line in IN:
                line = line.rstrip("\n")

                if line == "@<TRIPOS>MOLECULE":
                    mol_started = True

                if c == 4:
                    line = cm

                lines.append(line)

                if mol_started:
                    c += 1

        with open(output_file, "w") as OUT:
            OUT.write("\n".join(lines))



import warnings
import logging

from Bio.File import as_handle
from Bio.PDB.PDBExceptions import PDBConstructionException
from Bio.PDB.PDBExceptions import PDBConstructionWarning

from luna.pdb.parser.base import PDBParser
from luna.pdb.builder import StructureBuilder


logger = logging.getLogger()

PROBES_SET = ("ACD", "ACN", "ACT", "ADY", "AMN", "BDY", 
              "BEN", "BUT", "CHX", "DFO", "DME", "EOL", 
              "ETH", "PHN", "THS", "URE")


class FTMapParser(PDBParser):
    """
    Parser for FTMap PDB output.

    FTMap generates multiple `HEADER`-delimited chunks representing protein,
    clusters, and probe molecules. This parser merges them into a single Biopython
    Structure and performs:
      - optional filtering of clusters (`only_clusters`)
      - optional filtering of probes/residues (`only_compounds`)
      - serial number remapping
      - CONECT remapping
      - probe-to-ligand transformation
    """

    def __init__(self, 
                 PERMISSIVE=True, 
                 QUIET=False,
                 FIX_ATOM_NAME_CONFLICT=False,
                 FIX_OBABEL_FLAGS=False,
                 FIX_ALL_OBABEL_ERRORS=False):
        """
        Initialize an FTMapParser instance.

        Parameters
        ----------
        PERMISSIVE : bool
            Allow recovery from malformed PDB files.
            If False, exceptions in constructing the SMCRA data structure are fatal. 
            If true (DEFAULT), the exceptions are caught, but some residues or atoms 
            will be missing.
        QUIET : bool
            Suppress warnings during parsing.
        FIX_ATOM_NAME_CONFLICT : bool
            Fix conflicting atom names (e.g., hydrogens added by OpenBabel).
        FIX_OBABEL_FLAGS : bool
            Fix OpenBabel-added atom flags.
        FIX_ALL_OBABEL_ERRORS : bool
            Enable both atom name and flag corrections.
        """
        self.PERMISSIVE = bool(PERMISSIVE)
        self.QUIET = bool(QUIET)

        if FIX_ALL_OBABEL_ERRORS:
            FIX_ATOM_NAME_CONFLICT = True
            FIX_OBABEL_FLAGS = True
        
        # Correct conflicts in atom names.
        self.FIX_ATOM_NAME_CONFLICT = FIX_ATOM_NAME_CONFLICT
        # Do not fix empty chains using the default method. 
        # Instead, FTMapParser implements its own method to fix empty chains.
        self.FIX_EMPTY_CHAINS = False
        # Correct flags that were incorrectly set by OpenBabel.
        self.FIX_OBABEL_FLAGS = FIX_OBABEL_FLAGS
        # Correct all OpenBabel errors.
        self.FIX_ALL_OBABEL_ERRORS = FIX_ALL_OBABEL_ERRORS

    def get_structure(self, id, file, only_clusters=None, only_compounds=None):
        """
        Return a Biopython Structure object parsed from FTMap output.

        Parameters
        ----------
        id : str
            Structure ID.
        file : str or handle
            PDB file path or open handle.
        only_clusters : list of int, optional
            Limit parsing to specified FTMap cluster IDs.
        only_compounds : list of residue IDs, optional
            Limit output to specific residues.

        Returns
        -------
        Structure
        """
        with warnings.catch_warnings():
            if self.QUIET:
                warnings.filterwarnings("ignore", category=PDBConstructionWarning)

            with as_handle(file, mode='rU') as handle:
                self._parse_chunks(handle.readlines(), 
                                   id, 
                                   file,
                                   only_clusters=only_clusters,
                                   only_compounds=only_compounds)

            # Return the Structure instance
            return self.structure_builder.get_structure()

    def _parse_chunks(self, all_lines, id, file, only_clusters=None, only_compounds=None):
        """
        Parse all HEADER-separated chunks from an FTMap PDB file.

        This builds and merges partial structures using a temporary StructureBuilder.

        Parameters
        ----------
        lines : list of str
            Full PDB file lines.
        id : str
            Structure ID.
        file : str
            PDB file path.
        only_clusters : list of int, optional
        only_compounds : list of residue IDs, optional
        """
        chunks = []
        new_chunk = []

        # Track selected clusters/compounds
        self.only_clusters = set(only_clusters or [])
        self.only_compounds = self._get_valid_residue_ids(only_compounds or [])

        # Split file into chunks (on each HEADER)
        for line in lines:
            if line.startswith("HEADER"):
                if not line.startswith("HEADER    ") and line.startswith("HEADER "):
                    line = line.replace("HEADER ", "HEADER    ")
                if new_chunk:
                    chunks.append(new_chunk)
                    new_chunk = []
            new_chunk.append(line)

        if new_chunk:
            chunks.append(new_chunk)

        final_structure_builder = StructureBuilder()
        final_structure_builder.init_structure(id, file)
        final_structure_builder.init_model(0)

        curr_chain_id = "a"
        curr_serial_number = 1
        global_line_counter = 0

        for cid, chunk_lines in enumerate(chunks):
            self.structure_builder = StructureBuilder()
            self.structure_builder.init_structure(f"Chunk{cid}")

            self.header = None
            self.conects = None
            self.trailer = None
            self.line_counter = 0

            self._parse(chunk_lines)

            self.structure_builder.set_header(self.header)
            self.structure_builder.set_conects(self.conects)

            structure = self.structure_builder.get_structure()

            if structure.header["head"] == "protein":
                final_structure_builder.set_header(structure.header)
                if self.structure_builder.atom:
                    curr_serial_number = self.structure_builder.atom.serial_number + 1

            elif structure.header["head"].startswith("crosscluster.") and self.only_clusters:
                cluster_id = int(structure.header["head"].split(".")[1])
                if self.only_clusters and cluster_id not in self.only_clusters:
                    continue

            remapped_conects = {}
            for model_id in structure.child_dict:
                if not final_structure_builder.structure.has_id(model_id):
                    final_structure_builder.init_model(model_id)

                # Set the current model to be the one with id `model_id`, which can be an existing
                # model or a model just initialized.
                final_structure_builder.model = final_structure_builder.structure[model_id]

                empty_chains = []
                for chain_id in list(structure[model_id].child_dict):

                    # If some chain id is empty, then replace it with the current default chain id.
                    if chain_id.strip() == "":
                        structure[model_id][chain_id].id = curr_chain_id
                        chain_id = curr_chain_id
                        curr_chain_id = chr(ord(curr_chain_id) + 1)

                    if not final_structure_builder.model.has_id(chain_id):
                        final_structure_builder.model.add(structure[model_id][chain_id])

                        # Set the current chain to be the one given by the id `chain_id`.
                        final_structure_builder.chain = final_structure_builder.model[chain_id]

                        # Only modify residues if the structure contains a cluster of probes.
                        if structure.header["head"].startswith("crosscluster."):
                            remove_residues = []
                            # Loop over each residue and set probes to be considered ligands.
                            for res in final_structure_builder.chain:
                                # Modify probes to be hetatm.
                                if res.resname in PROBES_SET:
                                    res.id = ("H_%s" % res.resname, res.id[1], res.id[2])
                                    res.cluster_id = int(structure.header["head"].split(".")[1])

                                # Update the line number where the residue starts.
                                res.at_line = global_line_counter + res.at_line

                                # Ignore compounds that are not in the set ``only_compounds``.
                                if not self._accept_residue(res):
                                    remove_residues.append(res.id)
                                else:
                                    for atm in res.get_atoms():
                                        remapped_conects[atm.serial_number] = curr_serial_number
                                        atm.serial_number = curr_serial_number
                                        curr_serial_number += 1

                            # Remove residues not defined in ``only_compounds``.
                            for res_id in remove_residues:
                                del final_structure_builder.chain[res_id]
                    else:
                        # Set the current chain to be the one given by the id `chain_id`.
                        final_structure_builder.chain = final_structure_builder.model[chain_id]

                        # Loop over each residue and set probes to be considered ligands.
                        for res in structure[model_id][chain_id]:
                            # Only modify residues if the structure contains a cluster of probes.
                            if structure.header["head"].startswith("crosscluster."):
                                # Modify probes to be hetatm.
                                if res.resname in PROBES_SET:
                                    res.id = ("H_%s" % res.resname, res.id[1], res.id[2])
                                    res.cluster_id = int(structure.header["head"].split(".")[1])

                            # Update the line number where the residue starts.
                            res.at_line = global_line_counter + res.at_line

                            try:
                                # Ignore compounds that are not in the set ``only_compounds``.
                                if not self._accept_residue(res):
                                    continue
                                else:
                                    if curr_serial_number is not None:
                                        for atm in res.get_atoms():
                                            remapped_conects[atm.serial_number] = curr_serial_number
                                            atm.serial_number = curr_serial_number
                                            curr_serial_number += 1

                                final_structure_builder.chain.add(res)
                            except PDBConstructionException as message:
                                self._handle_PDB_exception(message, res.at_line)

                    if len(final_structure_builder.chain) == 0:
                        empty_chains.append(final_structure_builder.chain.id)

                # Remove empty chains.
                for chain_id in empty_chains:
                    del final_structure_builder.model[chain_id]

            # Remap the conects based on the atoms' new numbering.
            if remapped_conects:
                new_conects = {}
                for atm1, atoms in self.structure_builder.conects.items():
                    # Ignore atoms that were removed due to ``only_clusters`` and ``only_compounds``.
                    if atm1 not in remapped_conects:
                        continue

                    new_atoms_list = []
                    for atm2 in atoms:
                        # Ignore atoms that were removed due to ``only_clusters`` and ``only_compounds``.
                        if atm2 not in remapped_conects:
                            continue
                        new_atoms_list.append(remapped_conects[atm2])
                    new_conects[remapped_conects[atm1]] = new_atoms_list

                self.structure_builder.set_conects(new_conects)

                # Update the conects in final_structure_builder with the
                # remapped conects from the current structure
                final_structure_builder.conects.update(self.structure_builder.conects)
            elif structure.header["head"] == "protein":
                # Update the conects in final_structure_builder with the
                # conects from the original protein structure
                final_structure_builder.conects.update(self.structure_builder.conects)

            global_line_counter += self.structure_builder.line_counter

        self.structure_builder = final_structure_builder

    def _get_valid_residue_ids(self, res_ids):
        """
        Validate compound residue ID filters.

        Parameters
        ----------
        res_ids : iterable
            Residue ID tuples.

        Returns
        -------
        set
            Set of valid 3-tuple or 4-tuple residue IDs.
        """
        valid_residue_ids = set()

        for res_id in res_ids:
            if not isinstance(res_id, tuple) or len(res_id) not in [3, 4]:
                logger.warning("The residue id '%s' is not valid. "
                               "A valid residue id consists of a tuple containing three (hetflag, resseq, icode) or "
                               "four (structure id, model id, chain id, (hetflag, resseq, icode)) elements." % str(res_id))
                continue
            if len(res_id) == 4:
                if not isinstance(res_id[3], tuple) or len(res_id[3]) != 3:
                    logger.warning("The residue id '%s' is not valid. "
                                   "A valid residue id consists of a tuple containing three (hetflag, resseq, icode) or "
                                   "four (structure id, model id, chain id, (hetflag, resseq, icode)) elements." % str(res_id))
                    continue
            valid_residue_ids.add(res_id)
        return valid_residue_ids

    def _accept_residue(self, res):
        """
        Check whether a residue passes the compound filter.

        Parameters
        ----------
        res : Residue
            A Biopython Residue object.

        Returns
        -------
        bool
            True if accepted.
        """
        if self.only_compounds:
            for res_id in self.only_compounds:
                if len(res_id) == 4:
                    if res_id[1:] == res.get_full_id()[1:]:
                        return True
                elif len(res_id) == 3:
                    if res_id == res.id:
                        return True
            return False
        return True



from os import path

from luna.util.default_values import DOCK6_PATH

import logging
logger = logging.getLogger()


class Params(dict):

    def __init__(self, params=None):
        params = params or {}
        super().__init__(params)

    @property
    def names(self):
        """list: The list of parameters."""
        return sorted(self.keys())

    def save(self, output_file):
        # Don't create the output file if the params dictionary is empty.
        if not self:
            logger.info("There are no parameters to be saved.")
            return

        with open(output_file, "w") as OUT:
            for k, v in self.items():
                k = k.ljust(61)
                OUT.write("%s%s\n" % (k, str(v)))

    @classmethod
    def load(cls, input_file):
        params = {}

        with open(input_file) as IN:
            for line in IN:
                line = line.rstrip("\n")

                name, val = line.split(" ", 1)
                val = val.lstrip(" ")

                try:
                    val = int(val)
                except ValueError:
                    try:
                        val = float(val)
                    except ValueError:
                        pass
                params[name] = val

        return cls(params)


class DefaultParams(Params):

    def __init__(self, **kwargs):
        dock_path = path.abspath(path.join(path.realpath(__file__), '../'))
        default_file_path = f"{dock_path}/default.cfg"

        params = Params.load(default_file_path)

        for k, v in kwargs.items():
            params[k] = v

        super().__init__(params)



# Source: http://www.codekoala.com/posts/command-line-progress-bar-python/

from threading import Thread, Event

import time
import sys
from colorlog.escape_codes import escape_codes, parse_colors


import logging
logger = logging.getLogger()


class ProgressData:
    """
    Custom data structure to store data generated during the
    execution of a single task.

    Parameters
    ----------
    input_data : any
        The input data of the executed task.
    proc_time : float
        How long a task took to be executed.
    output_data : any, optional
        The output data produced by a given task.
        The data type is the same as the executed function's return.
    exception : :py:class:`Exception`, optional
        If an exception was raised, then ``exception`` stores an
        Exception object. Otherwise, ``exception`` will be set to None.
    func : function, optional
        The executed function for reference.
    """

    def __init__(self, input_data, proc_time,
                 output_data=None, exception=None, func=None):

        self.input_data = input_data
        self.output_data = output_data
        self.proc_time = proc_time
        self.exception = exception
        self.func = func


class ProgressResult:
    """
    Custom iterable class to store `ProgressData` objects as they are produced
    during the execution of a set of tasks.

    This class implements append() and __iter__() by default.

    Parameters
    ----------
    results : list, optional
        A pre-populated list of `ProgressData` objects.

    Attributes
    ----------
    results : list
        The list of `ProgressData` objects.
    """

    def __init__(self, results=None):
        if results and not isinstance(results, list):
            raise TypeError("A list was expected but a different data type "
                            "was provided.")
        self.results = results or []

    @property
    def inputs(self):
        """list: Inputs from each `ProgressData` object stored in \
        ``results``."""
        try:
            return [r.input_data for r in self.results]
        except AttributeError:
            raise TypeError("An invalid object was found in 'results'. "
                            "Only ProgressData objects are valid.")

    @property
    def outputs(self):
        """list of tuple: Outputs from each `ProgressData` object stored in
        ``results``. Each tuple contains the input and the output produced for
        that input."""
        try:
            return [(r.input_data, r.output_data) for r in self.results]
        except AttributeError:
            raise TypeError("An invalid object was found in 'results'. "
                            "Only ProgressData objects are valid.")

    @property
    def errors(self):
        """list of tuple: Errors from each `ProgressData` object stored in \
        ``results``.
        Each tuple contains the input and the exception raised during the
        execution of a task with that input."""
        try:
            return [(r.input_data, r.exception) for r in self.results
                    if r.exception is not None]
        except AttributeError:
            raise TypeError("An invalid object was found in 'results'. "
                            "Only ProgressData objects are valid.")

    def append(self, r):
        """Add a new `ProgressData` object to ``results``"""
        if not isinstance(r, ProgressData):
            raise TypeError("Only ProgressData objects are valid.")
        self.results.append(r)

    def __len__(self):
        return len(self.results)

    def __iter__(self):
        for r in self.results:
            yield r


class ProgressTracker:
    """
    A progress tracker for tasks queued in ``queue``.

    Parameters
    ----------
    ntasks : int
        The number of tasks to be executed.
    queue : :py:class:`~multiprocessing.Queue`
        A queue to track the tasks' progress.
    task_name : str, optional
        A name to identify the set of tasks.

    Attributes
    ----------
    ntasks : int
        The number of tasks to be executed.
    queue : :py:class:`~multiprocessing.Queue`
        The queue to track the tasks' progress from.
    task_name : str, optional
        The name to identify the set of tasks.
    results : ProgressResult
        Store results and any errors found during the tasks' processing.
    nerrors : int
        Number of errors.
    """

    def __init__(self, ntasks, queue, task_name=None):
        self.ntasks = ntasks
        self.task_name = task_name

        self._progress = 0

        self.queue = queue   # used to communicate progress to the thread
        self._event = Event()    # used to tell the thread when to finish
        self._progress_bar = Thread(target=self._print_progress,
                                    args=(self._event, self.queue))
        self._progress_bar.daemon = True

        # Save results and any errors found during the task processing.
        self.results = ProgressResult()
        self.nerrors = 0

        self._running_times = []
        self._start_time = None
        self._end_time = None

    def _show_progress_bar(self, p, perc):
        task_name = ""
        if self.task_name:
            task_name = " - %s" % self.task_name

        msg = ('%s%% [%s] %d/%d [Avg: %.2fs/task; Errors: %d]%s.'
               % (int(perc), ("\u25A0" * int(perc / 2)).ljust(50, ' '),
                  p, self.ntasks, self.avg_running_time,
                  self.nerrors, task_name))

        format_str = '\r[%s]    %s%s %s%s  %s'
        progress_str = \
            (format_str % (time.strftime('%Y-%m-%d %H:%M:%S'),
             parse_colors("purple"),
             "PROGRESS".ljust(10, " "),
             escape_codes["reset"],
             "".rjust(26, " "),
             msg))

        sys.stdout.write(progress_str)
        sys.stdout.flush()

    def _print_progress(self, e, q):
        """Updates a progress bar on stdout anytime progress is made"""

        while True:
            while not q.full():
                # Stops the loop if function end() is called before the queue
                # gets full.
                if e.is_set():
                    break
                # wait for more progress to be made
                time.sleep(0.1)

            # If our event is set and there is any progress in the queue,
            # break out of the infinite loop and prepare to terminate this
            # thread
            if e.is_set() and q.full() is False:
                break

            # get the current progress data.
            progress_data = q.get()

            if progress_data is not None:
                self.results.append(progress_data)

                self._running_times.append(progress_data.proc_time)

                if progress_data.exception is not None:
                    self.nerrors += 1

                self.progress += 1

            perc = round((self.progress / self.ntasks), 2) * 100 if self.ntasks > 0 else 0
            self._show_progress_bar(self.progress, perc)

    @property
    def progress(self):
        """int: Current number of tasks executed."""
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value

    @property
    def running_time(self):
        """float: Total running time."""
        if self._start_time and self._end_time:
            return round(self._end_time - self._start_time, 2)
        return None

    @property
    def avg_running_time(self):
        """float: Average running time."""
        if self._running_times:
            total = sum(self._running_times)
            count = len(self._running_times)
            return round(total / count, 2)
        else:
            return 0

    def start(self):
        """ Start the progress tracker. """
        self._start_time = round(time.time(), 2)
        self._progress_bar.start()

    def end(self):
        """ Finish the progress tracker. """
        self._event.set()
        self._progress_bar.join()
        self._end_time = round(time.time(), 2)
        sys.stdout.write('\n')



from subprocess import Popen, PIPE

from luna.wrappers.chemaxon.util import check_license
from luna.wrappers.base import MolWrapper
from luna.util.exceptions import ProcessingFailed

import logging
logger = logging.getLogger()


class StereoisomerGenerator:

    def __init__(self,
                 max_stereoisomers=1000,
                 protect_double_bond_stereo=False,
                 protect_tetrahedral_stereo=False,
                 verify_3d=False,
                 mol_obj_type="rdkit",
                 cxcalc="cxcalc"):

        self.max_stereoisomers = max_stereoisomers
        self.protect_double_bond_stereo = protect_double_bond_stereo
        self.protect_tetrahedral_stereo = protect_tetrahedral_stereo
        self.verify_3d = verify_3d

        self.mol_obj_type = mol_obj_type
        self.cxcalc = cxcalc

    def _prep_opts(self):

        opt_list = []

        # Maximum number of stereoisomers to be generated
        opt_list += ["-m", str(self.max_stereoisomers)]

        # Protect aromaticity
        opt_list += ["-D", str(self.protect_double_bond_stereo)]

        # Protect charge
        opt_list += ["-T", str(self.protect_tetrahedral_stereo)]

        # Exclude antiaromatic compounds
        opt_list += ["-v", str(self.verify_3d)]

        # SMILES output.
        opt_list += ["-f", "smiles"]

        return opt_list

    def _run_cxcalc(self, mol_input):
        opt_list = self._prep_opts()

        self.mol_input = mol_input

        plugin = "stereoisomers"
        args = [self.cxcalc] + [mol_input] + [plugin] + opt_list

        p = Popen(args, stdout=PIPE, stderr=PIPE)
        try:
            stdout, stderr = p.communicate()
        except Exception:
            p.kill()
            raise

        error_lines = stderr.decode().strip().split("\n")
        if len(error_lines) == 1 and error_lines[0] == "":
            logger.debug("Stereoisomers for input '%s' created with success."
                         % mol_input)
        else:
            logger.error(stderr.decode())
            raise ProcessingFailed("cxcalc could not generate stereoisomers. "
                                   "Check the logs for more information.")

        self.stereoisomers = []
        for smi in stdout.decode().strip().split("\n"):
            smi = smi.strip()

            if smi == "":
                continue

            mol_obj = MolWrapper.from_smiles(smi)

            mol_id = "Stereoisomer %d" % len(self.stereoisomers)
            mol_obj.set_name(mol_id)

            self.stereoisomers.append(mol_obj)

    def generate(self, mol_input):

        self.stereoisomers = []

        check_license()

        self._run_cxcalc(mol_input)

        return self.stereoisomers



import numpy as np
from Bio.PDB.kdtrees import KDTree


class AtomGroupNeighborhood:
    """ Class for fast neighbor atom groups searching.

    ``AtomGroupNeighborhood`` makes use of a KD Tree implemented in C,
    so it's fast.

    Parameters
    ----------
    atm_grps : iterable of `AtomGroup`, optional
        A sequence of `AtomGroup` objects, which is used in the queries.
        It can contain atom groups from different molecules.
    bucket_size : int
        Bucket size of KD tree.
        You can play around with this to optimize speed if you feel like it.
        The default value is 10.

    """

    def __init__(self, atm_grps, bucket_size=10):
        self.atm_grps = list(atm_grps)

        # get the coordinates
        coord_list = [ga.centroid for ga in self.atm_grps]

        # to Nx3 array of type float
        self.coords = np.array(coord_list).astype("d")
        assert(bucket_size > 1)
        assert(self.coords.shape[1] == 3)
        self.kdt = KDTree(self.coords, bucket_size)    

    def search(self, center, radius):
        """Return all atom groups in ``atm_grps`` that is up to a maximum of
        ``radius`` away (measured in Å) of ``center``.

        For atom groups with more than one atom, their centroid is used as a
        reference.
        """
        center = np.require(center, dtype="d", requirements="C")
        if center.shape != (3,):
            raise Exception("Expected a 3-dimensional NumPy array")
        points = self.kdt.search(center, radius)            
        n_grps_list = [self.atm_grps[point.index] for point in points]

        return n_grps_list



from Bio.PDB.Chain import Chain as BioChain

from luna.pdb.core.entity import Entity


class Chain(BioChain, Entity):
    """
    Custom Chain subclass with support for SMCRA traversal
    and reference flagging for processing.

    Parameters
    ----------
    id : str
        Chain identifier (e.g., "A").
    """

    def __init__(self, id):
        super().__init__(id)
        
        # By default: no chain is a reference for processing.
        self._is_reference = False

    def __repr__(self):
        return f"<Chain (LUNA subclass) id={self.id}>"

    def __lt__(self, other):
        return self.id < other.id

    def is_reference(self):
        """
        Return True if this entity is marked as the reference for interaction analysis.

        This flag indicates that the user selected this entity
        as the *main subject* of interaction evaluation — interactions will be computed
        between this entity and others nearby (e.g., ligands, water, ions).

        Returns
        -------
        bool
            True if this entity is marked for interaction reference.
        """
        return self._is_reference

    def set_as_reference(self, is_reference=True):
        """
        Mark or unmark the chain and all its residues as references.

        This method sets a flag indicating that this chain should be 
        treated as the main subject for calculating interactions 
        with nearby molecules.

        Parameters
        ----------
        is_reference : bool, optional
            If True, mark the entity as the interaction reference.
            If False, unmark it. Defaults to True.

        See Also
        --------
        is_reference : Check whether this entity is currently marked.
        """
        self._is_reference = is_reference
        for res in self.get_residues():
            if hasattr(res, "set_as_reference"):
                res.set_as_reference(is_reference)



from luna.mol.entry import MolFileEntry
from luna.wrappers.pymol import PymolWrapper, PymolSessionManager
from luna.pdb.io.helpers import entity_to_string
from luna.util.exceptions import PymolSessionNotInitialized
from luna.util.file import is_file_valid


import logging

logger = logging.getLogger()


class ShellViewer(PymolSessionManager):
    """Class that inherits from
    :class:`~luna.wrappers.pymol.PymolSessionManager` and implements
    :meth:`set_view` to depict shells (IFP features) and interactions
    in Pymol and save the view as a Pymol session.

    This class can be used to visualize multiple complexes into the same Pymol
    session, for instance, to compare binding modes and analyze similar shells.
    To do so, it is recommended that the protein structures are in the same
    coordinate system, i.e., they should be aligned first. This can be achieved
    with :func:`luna.align.tmalign.align_2struct` or any other tool of
    your preference.

    Examples
    --------

    In the below example, we will assume a LUNA project object named
    ``proj_obj`` already exists.

    To visualize shells, we first need to generate them. So, let's define a
    :class:`~luna.interaction.fp.shell.ShellGenerator` object that will create
    shells over 2 iterations (levels). At each iteration, the shell radius will
    be increased by 3 and substructural information will be encoded following
    EIFP definition. Here, as an example, we will generate shells for the first
    :class:`~luna.mol.groups.AtomGroupsManager` object at ``proj_obj``.

    >>> from luna.interaction.fp.shell import ShellGenerator
    >>> from luna.interaction.fp.type import IFPType
    >>> num_levels, radius_step = 2, 3
    >>> sg = ShellGenerator(num_levels, radius_step, ifp_type=IFPType.EIFP)
    >>> atm_grps_mngr = list(proj_obj.atm_grps_mngrs)[0]
    >>> sm = sg.create_shells(atm_grps_mngr)

    The function
    :meth:`~luna.interaction.fp.shell.ShellGenerator.create_shells` returns a
    :class:`~luna.interaction.fp.shell.ShellManager` object, which provides
    built-in methods to access the created shells. Therefore, you can interact
    with it to select the shells you want to visualize in Pymol. As an example,
    let's select all unique shells at the last level:

    >>> shells = sm.get_shells_by_level(num_levels - 1, unique_shells=True)

    .. note::
        Levels are 0-indexed. So, the first level is 0, second is 1, etc.
        That means if ``num_levels`` is 5, the last level will be 4.

    As `ShellViewer` expects a list of tuples, we will define it now. The first
    item of the tuple is the :class:`~luna.mol.entry.Entry` instance, which
    represents a ligand. This can be obtained directly from the
    :class:`~luna.mol.groups.AtomGroupsManager` object. The second item is the
    list of shells you want to visualize. Finally, the third item can be either
    a PDB file or a directory. In this example, we will use the PDB directory
    defined during the LUNA project initialization.

    >>> shell_tuples = [(atm_grps_mngr.entry, shells, proj_obj.pdb_path)]

    To finish, we now create a new `ShellViewer` object and call
    :meth:`~luna.wrappers.pymol.PymolSessionManager.new_session`, which will
    initialize the session, depict shells and interactions, and save the
    session to an output PSE file.

    >>> from luna.interaction.fp.view import ShellViewer
    >>> sv = ShellViewer()
    >>> sv.new_session(shell_tuples, "example.pse")
    """

    def set_view(self, shell_tuples):
        """Depict shells (IFP features) and interactions into the current
        Pymol session.

        Parameters
        ----------
        shell_tuples : iterable of tuple
            Each tuple must contain three items: an
            :class:`~luna.mol.entry.Entry` instance,
            an iterable of :class:`~luna.interaction.fp.shell.Shell`,
            and a PDB file or the directory where the PDB file is located.
        """

        if not isinstance(self.wrapper, PymolWrapper):
            raise PymolSessionNotInitialized("No session was initialized.")

        for target_entry, shells, pathname in shell_tuples:

            if is_file_valid(pathname):
                pdb_file = pathname
            else:
                pdb_file = "%s/%s.pdb" % (pathname, target_entry.pdb_id)

            main_grp = target_entry.to_string(sep="-")

            mol_block = \
                (entity_to_string(target_entry.get_biopython_structure())
                 if isinstance(target_entry, MolFileEntry) else None)

            # Load PDB and extract hetatm.
            self.load_pdb(pdb_file, main_grp, mol_block)

            residue_selections = set()
            for index, shell in enumerate(shells):
                sphere_obj = ("%s.spheres.s%d_lvl%d_center%d"
                              % (main_grp, index, shell.level,
                                 hash(shell.central_atm_grp)))
                centroid_obj = "%s.centroid" % (sphere_obj)

                centroids = list(shell.central_atm_grp.centroid)
                self.wrapper.add_pseudoatom(centroid_obj,
                                            {"color": "white",
                                             "pos": centroids})
                self.wrapper.hide([("nonbonded", centroid_obj)])
                self.wrapper.show([("spheres", centroid_obj),
                                   ("nb_spheres", centroid_obj),
                                   ("dots", centroid_obj)])
                self.wrapper.set("dot_color", "red")
                self.wrapper.set("sphere_scale",
                                 shell.radius,
                                 {"selection": centroid_obj})
                self.wrapper.set("sphere_transparency", 0.85,
                                 {"selection": centroid_obj})
                self.wrapper.run_cmds([("center",
                                        {"selection": centroid_obj})])

                atm_names = [a.name
                             for a in sorted(shell.central_atm_grp.atoms)]
                self.wrapper.label([(centroid_obj,
                                     '"%s"' % "+".join(atm_names))])

                # Add interactions and styles.
                interacting_residue_sels = \
                    self.set_interactions_view(shell.interactions,
                                               main_grp, sphere_obj)

                residue_selections.update(interacting_residue_sels)

            if residue_selections:
                self.wrapper.select(name="%s.inter_residues" % main_grp,
                                    selection=" or ".join(residue_selections))

        self.set_last_details_to_view()



from Bio.PDB.Atom import Atom as BioAtom
from Bio.PDB.Atom import DisorderedAtom as BioDisorderedAtom


class Atom(BioAtom):
    """
    A custom Atom class that supports hierarchical traversal and
    additional coordination metadata.

    Inherits from Biopython's Atom and adds:
    - `metal_coordination` set
    - `hierarchy_name` property
    - `get_parent_by_level()` for SMCRA traversal
    - Custom sorting via __lt__

    Parameters
    ----------
    name : str
        Atom name (e.g., "CA").
    coord : array_like
        Cartesian coordinates.
    bfactor : float
        B-factor value.
    occupancy : float
        Occupancy.
    altloc : str
        Alternate location indicator.
    fullname : str
        Full atom name with padding (e.g., " CA ").
    serial_number : int
        Atom serial number.
    element : str, optional
        Atomic element.
    """
    def __init__(self, name, coord, bfactor, occupancy, altloc,
                 fullname, serial_number, element=None):
        
        super().__init__(name, coord, bfactor, occupancy, altloc,
                         fullname, serial_number, element)
        self.metal_coordination = set()

    def __repr__(self):
        return f"<Atom (LUNA subclass) {self.get_id()}>"

    def __lt__(self, other):
        """
        Atom sorting based on structure hierarchy and residue index.

        Ensures original PDB order is preserved.
        """
        id1 = self.get_full_id()[0:3] + (self.get_parent().idx,) + self.get_full_id()[4:]
        id2 = other.get_full_id()[0:3] + (other.get_parent().idx,) + other.get_full_id()[4:]
        return id1 < id2

    @property
    def hierarchy_name(self):
        """
        Return full hierarchical name of the atom.
    
        Format: structure/model/chain/resname/resid/atom[-altloc]
        """
        full_id = self.get_full_id()
        parent = self.get_parent()
    
        parent_path = '/'.join(str(x) for x in full_id[0:3])
        resname = f"{parent.resname}/{parent.id[1]}{parent.id[2].strip()}"
        atom_name = self.name + (f"-{self.altloc}" if self.altloc != " " else "")
    
        return f"{parent_path}/{resname}/{atom_name}"

    def has_metal_coordination(self) -> bool:
        """Return True if the atom is metal-coordinated."""
        return bool(self.metal_coordination)

    # As the Atom class only partly implements the Entity interface, 
    # this function was included in both classes.
    def get_parent_by_level(self, level):
        """
        Return the parent entity at the specified SMCRA level.

        Parameters
        ----------
        level : str
            One of "A", "R", "C", "M", "S".

        Returns
        -------
        Entity
            The parent entity at the requested level.
        """
        if self.level == level:
            return self
        else:
            return self.parent.get_parent_by_level(level)


class DisorderedAtom(BioDisorderedAtom):
    """
    LUNA's DisorderedAtom placeholder.

    Currently identical to Biopython's, but allows future extensions and consistent imports.
    """
    def __init__(self, id):
        super().__init__(id)

    def __repr__(self):
        """Return disordered atom identifier."""
        if self.child_dict:
            return f"<DisorderedAtom (LUNA subclass) {self.get_id()}>"
        else:
            return f"<Empty DisorderedAtom (LUNA subclass) {self.get_id()}>"



from itertools import chain
from collections import defaultdict
from enum import Enum
import numpy as np
import mmh3

from luna.version import __version__
from luna.util.exceptions import ShellCenterNotFound
from luna.util.default_values import CHEMICAL_FEATURE_IDS, INTERACTION_IDS
from luna.interaction.fp.fingerprint import (DEFAULT_FP_LENGTH, Fingerprint,
                                             CountFingerprint)
from luna.interaction.fp.type import IFPType
from luna.mol.groups import PseudoAtomGroup
from luna.pdb.neighbors.atom_group import AtomGroupNeighborhood
from luna.mol.features import ChemicalFeature


import logging

logger = logging.getLogger()


PAD_DEFAULT = -9


class CompoundClassIds(Enum):
    """An enumeration of compound classes."""
    HETATM = 1
    RESIDUE = 2
    NUCLEOTIDE = 3
    WATER = 4
    UNKNOWN = 5


class ShellManager:

    """Store and manage :class:`Shell` objects.

    Parameters
    ----------
    num_levels : int
        The maximum number of iterations for fingerprint generation.
    radius_step : float
        The multiplier used to increase shell size at each iteration.
        At iteration 0, shell radius is 0 * ``radius_step``, at iteration 1,
        radius is 1 * ``radius_step``, etc.
    fp_length : int
        The fingerprint length (total number of bits).
    ifp_type : :class:`~luna.interaction.fp.type.IFPType`
        The fingerprint type (EIFP, FIFP, or HIFP).
    shells : iterable of :class:`Shell`, optional
        An initial sequence of :class:`Shell` objects (fingerprint features).
    verbose : bool
        If True, warnings issued during the usage of this `ShellManager` will
        be displayed. The default value is False.

    Attributes
    ----------
    num_levels : int
        The maximum number of iterations for fingerprint generation.
    radius_step : float
        The multiplier used to increase shell size at each iteration.
    fp_length : int
        The fingerprint length (total number of bits).
    ifp_type : :class:`~luna.interaction.fp.type.IFPType`
        The fingerprint type (EIFP, FIFP, or HIFP).
    shells : iterable of :class:`Shell`
        The sequence of shells (fingerprint features).
    verbose : bool
        The verbosity state.
    version : str
        The LUNA's version with which shells were generated.
    levels : dict of {int: list of `Shell`}
        Register shells by level, where keys are levels and values are
        lists of `Shell` objects.

        .. note::
            Levels are 0-indexed. So, the first level is 0, second is 1, etc.
            That means if ``num_levels`` is 5, the last level will be 4.
    centers : dict of dict of {int: `Shell`}
        Register shells by center, where keys are \
        :class:`~luna.mol.groups.AtomGroup` objects and values are dict that
        store all shells generated for that center at each iteration (level).
    """

    def __init__(self,
                 num_levels,
                 radius_step,
                 fp_length,
                 ifp_type,
                 shells=None,
                 verbose=False):
        self.num_levels = num_levels
        self.radius_step = radius_step
        self.fp_length = fp_length
        self.ifp_type = ifp_type

        self.shells = shells or []
        self.verbose = verbose

        self.version = __version__

        self._init_controllers()

    @property
    def num_shells(self):
        """int, read-only: Total number of shells in ``shells``."""
        return len(self.shells)

    @property
    def num_unique_shells(self):
        """int, read-only: Total number of unique shells in ``shells``."""
        return len(self.unique_shells)

    @property
    def unique_shells(self):
        """iterable of `Shell`, read-only: Unique shells. \
        Return the same as :meth:`get_valid_shells`."""
        return self.get_valid_shells()

    def find_similar_shell(self, shell):
        """Find a shell in ``shells`` similar to ``shell``.

        Two shells are similar if they represent the same substructural
        information.

        Parameters
        ----------
        shell : `Shell`

        Returns
        -------
         : `Shell` or None
            Return a similar shell or None if it does not find any.
        """
        for added_shell in self.shells:
            # For each group of similar shells, it will exist only
            # one valid shell.
            if added_shell.is_valid():
                if shell.is_similar(added_shell):
                    return added_shell
        return None

    def add_shell(self, shell):
        """Add a new shell to ``shells``.

        Parameters
        ----------
        shell : `Shell`
        """

        # Any new shell without interactions from level 1 onwards will be
        # automatically considered invalid.
        if shell.level > 0 and len(shell.interactions) == 0:
            shell.valid = False
        else:
            found_shell = self.find_similar_shell(shell)

            if found_shell:
                if found_shell.level < shell.level:
                    shell.valid = False

                elif (found_shell.level == shell.level
                        and found_shell.identifier <= shell.identifier):
                    shell.valid = False

                else:
                    found_shell.valid = False

        self.shells.append(shell)
        self.levels[shell.level].append(shell)
        self.centers[shell.central_atm_grp][shell.level] = shell

    def get_valid_shells(self):
        """Return only valid shells.

        A shell is considered invalid if, by the time it is added in
        ``shells``, there is another shell representing the same
        substructural information. That means this shell is not unique
        and does not contributes to any new information.

        On the other hand, if the shell contributes by adding new information
        to ``shells``, then it will be considered valid and unique. So, the
        first shell of a series of shells containing the same information is
        considered valid and the others invalid.

        Returns
        -------
         : list of `Shell`
        """
        return [s for s in self.shells if s.is_valid()]

    def get_shells_by_identifier(self, identifier, unique_shells=False):
        """Get shells by identifier.

        Parameters
        ----------
        identifier : int
            The shell identifier.
        unique_shells : bool
            If True, return only unique shells.
            Otherwise, return all shells having the identifier
            ``identifier`` (the default).

        Returns
        -------
         : list of `Shell`
        """
        if unique_shells:
            return [s for s in self.shells
                    if s.identifier == identifier and s.is_valid()]
        else:
            return [s for s in self.shells if s.identifier == identifier]

    def get_shells_by_level(self, level, unique_shells=False):
        """Get shells by level (iteration number).

        Parameters
        ----------
        level : int
        unique_shells : bool
            If True, return only unique shells.
            Otherwise, return all shells at level ``level`` (the default).

        Returns
        -------
         : list of `Shell`
        """
        shells = []

        if level in self.levels:
            if unique_shells:
                shells = [s for s in self.levels[level] if s.is_valid()]
            else:
                shells = self.levels[level]
        elif self.verbose:
            logger.warning("The informed level '%d' does not exist." % level)

        return shells

    def get_shells_by_center(self, center, unique_shells=False):
        """Get shells by center (:class:`~luna.mol.groups.AtomGroup` object).

        Parameters
        ----------
        center : :class:`~luna.mol.groups.AtomGroup`
            The center of a shell, which consists of an
            :class:`~luna.mol.groups.AtomGroup` object.
        unique_shells : bool
            If True, return only unique shells.
            Otherwise, return all shells generated for center
            ``center`` (the default).


        Returns
        -------
         : dict of {int: `Shell`}
            All shells generated for center ``center`` at each iteration (key).
        """
        shells = {}

        if center in self.centers:
            if unique_shells:
                shells = {i: s for i, s in self.centers[center].items()
                          if s.is_valid()}
            else:
                shells = self.centers[center]
        elif self.verbose:
            logger.warning("The informed center '%s' does not exist." % center)

        return shells

    def get_shell_by_center_and_level(self,
                                      center,
                                      level,
                                      unique_shells=False):
        """Get the shell generated for center ``center``
        (:class:`~luna.mol.groups.AtomGroup` object) at level (iteration)
        ``level``.

        Parameters
        ----------
        center : :class:`~luna.mol.groups.AtomGroup`
            The center of a shell, which consists of an
            :class:`~luna.mol.groups.AtomGroup` object.
        level : int
            The target level (iteration).
        unique_shells : bool
            If True, return the `Shell` object if it is unique and None
            otherwise. The default value is False.

        Returns
        -------
         : `Shell` or None
            The shell generated for center ``center`` at level ``level``.
            If the `Shell` object is not unique, return None.
        """
        shell = self.centers.get(center, {}).get(level)

        if shell is None and self.verbose:
            logger.warning("The informed center '%s' does not exist in "
                           "the level '%d'." % (center, level))

        if unique_shells:
            shell = shell if shell.is_valid() else None
            if shell is None and self.verbose:
                logger.warning("The shell found with center '%s' and level "
                               "'%d' is not unique." % (center, level))

        return shell

    def get_previous_shell(self, center, curr_level, unique_shells=False):
        """Get the last shell having center ``center`` that was generated
        before level ``curr_level``. For instance, if the current level
        (iteration) is 5 and the last valid shell generated for center
        :math:`C` was at level 4, then :meth:`get_previous_shell` would
        return that shell at level 4.

        Parameters
        ----------
        center : :class:`~luna.mol.groups.AtomGroup`
            The center of a shell, which consists of an
            :class:`~luna.mol.groups.AtomGroup` object.
        curr_level : int
            The current level (iteration).
        unique_shells : bool
            If True, ignore non-valid shells and go down to inferior
            levels until a valid shell is found. If level 0 was reached
            and no valid shell was found, then return None.
            The default value is False.

        Returns
        -------
         : `Shell` or None
            The first previous valid shell or None if no valid shell was found.
        """
        shell = None
        while curr_level != 0 and shell is None:
            level = curr_level - 1
            shell = self.get_shell_by_center_and_level(center,
                                                       level,
                                                       unique_shells)
            curr_level = level

        if shell is None and self.verbose:
            logger.warning("No previous shell centered on '%s' departing from "
                           "the level '%d' was found." % (center, level))

        return shell

    def get_last_shell(self, center, unique_shells=False):
        """Get the last shell generated for center ``center``.

        Parameters
        ----------
        center : :class:`~luna.mol.groups.AtomGroup`
            The center of a shell, which consists of an
            :class:`~luna.mol.groups.AtomGroup` object.
        unique_shells : bool
            If True, ignore non-valid shells.
            That means shells generated at superior levels may be ignored
            if they are not valid. The default value is False.

        Returns
        -------
         : `Shell` or None
            The last shell generated for center ``center`` or None if
            no valid shell was found.
        """
        shell = None

        shells = self.get_shells_by_center(center, unique_shells)
        if shells:
            # Sort by key and get a shell based on the latest value.
            shell = shells[sorted(shells, key=int)[-1]]
        elif self.verbose:
            logger.warning("No shell centered on '%s' was found." % center)

        return shell

    def get_identifiers(self, level=None, unique_shells=False):
        """Get all shells' identifier.

        Parameters
        ----------
        level : int, optional
            If provided, only return identifiers of shells at level ``level``.
        unique_shells : bool
            If True, ignore identifiers of non-valid shells.
            The default value is False.

        Returns
        -------
         : list of int
        """
        if level is not None:
            if unique_shells:
                identifiers = [s.identifier
                               for s in self.get_shells_by_level(level)
                               if s.is_valid()]
            else:
                identifiers = [s.identifier
                               for s in self.get_shells_by_level(level)]
        else:
            if unique_shells:
                identifiers = [s.identifier for s in self.shells
                               if s.is_valid()]
            else:
                identifiers = [s.identifier for s in self.shells]

        return sorted(identifiers)

    def to_fingerprint(self,
                       fold_to_length=None,
                       count_fp=False,
                       unique_shells=False):
        """Encode shells into an interaction fingerprint.

        Parameters
        ----------
        fold_to_length : int, optional
            If provided, fold the fingerprint to length ``fold_to_length``.
        count_fp : bool
            If True, create a count fingerprint
            (:class:`~luna.interaction.fp.fingerprint.CountFingerprint`).
            Otherwise, return a bit fingerprint
            (:class:`~luna.interaction.fp.fingerprint.Fingerprint`).
        unique_shells : bool
            If True, only unique shells are used to create the fingerprint.
            The default value is False.

        Returns
        -------
         : :class:`~luna.interaction.fp.fingerprint.CountFingerprint` or \
                :class:`~luna.interaction.fp.fingerprint.Fingerprint`
        """
        indices = self.get_identifiers(unique_shells=unique_shells)
        props = {"num_levels": self.num_levels,
                 "fp_length": self.fp_length,
                 "radius_step": self.radius_step}

        if count_fp:
            fp = CountFingerprint(indices,
                                  fp_length=self.fp_length,
                                  props=props)
        else:
            fp = Fingerprint(indices,
                             fp_length=self.fp_length,
                             props=props)

        if fold_to_length:
            return fp.fold(fold_to_length)
        return fp

    def trace_back_feature(self,
                           feature_id, ifp,
                           unique_shells=False):
        """Trace a feature from a fingerprint back to the shells
        that originated that feature.

        .. note::
            Due to fingerprint folding, multiple substructures may end up
            encoded in the same bit, the so-called collision problem.
            So, if the provided feature contains collisions, shells
            representing different substructures may be returned by
            :meth:`trace_back_feature`.

        Parameters
        ----------
        feature_id: int
            The target feature id.
        ifp : `Fingerprint`
            The fingerprint containing the feature ``feature_id``.
        unique_shells : bool
            If True, ignore identifiers of non-valid shells.
            The default value is False.

        Yields
        -------
         : `Shell`

        Examples
        --------

        In the below example, we will assume a LUNA project object named
        ``proj_obj`` already exists. Then, we will generate an EIFP
        fingerprint for the first :class:`~luna.mol.groups.AtomGroupsManager`
        object at ``proj_obj``.

        >>> from luna.interaction.fp.shell import ShellGenerator
        >>> from luna.interaction.fp.type import IFPType
        >>> atm_grps_mngr = list(proj_obj.atm_grps_mngrs)[0]
        >>> num_levels, radius_step = 2, 3
        >>> sg = ShellGenerator(num_levels, radius_step, ifp_type=IFPType.EIFP)
        >>> sm = sg.create_shells(atm_grps_mngr)
        >>> fp = sm.to_fingerprint(fold_to_length=1024, count_fp=True)
        >>> print(fp.indices)
        [   2   19   22   23   34   37   39   45   54   67   71   75   83   84
           93  109  138  140  157  162  181  182  186  187  191  194  206  209
          211  237  246  251  263  271  281  296  304  315  323  358  370  374
          388  392  399  400  419  439  476  481  487  509  519  527  532  578
          587  592  604  605  629  635  645  661  668  698  711  713  732  736
          740  753  764  795  813  815  820  824  825  831  836  855  873  882
          911  926  967  975  976  984  990  996 1020]

        Now, we can trace features back to original identifiers and investigate
        its substructural information.

        >>> ori_indices = list(sm.trace_back_feature(34, fp, \
unique_shells=True))
        >>> print(ori_indices)
        [(494318626, [<Shell: level=0, radius=0.000000, center=<AtomGroup: \
[<ExtendedAtom: 3QQK/0/A/GLN/85/CD>, <ExtendedAtom: 3QQK/0/A/GLN/85/NE2>, \
<ExtendedAtom: 3QQK/0/A/GLN/85/OE1>]>, interactions=0>])]

        """
        for ori_feature in ifp.unfolding_map[feature_id]:
            yield (ori_feature,
                   self.get_shells_by_identifier(ori_feature, unique_shells))

    def _init_controllers(self):
        levels = defaultdict(list)
        centers = defaultdict(lambda: defaultdict(lambda: None))

        for shell in self.shells:
            levels[shell.level].append(shell)
            centers[shell.central_atm_grp][shell.level] = shell

        self.levels = levels
        self.centers = centers


class Shell:

    """ A container to store substructural information, which is the base for
    LUNA fingerprints.

    Shells are centered on an atom or atom group
    (:class:`~luna.mol.groups.AtomGroup` objects) and represent all atoms and
    interactions explicitly within it.

    Parameters
    ----------
    central_atm_grp : :class:`~luna.mol.groups.AtomGroup`
        The shell center.
    level : int
        The level (iteration) at which the shell was generated.
    radius : float
        The shell radius.
    neighborhood : iterable of :class:`~luna.mol.groups.AtomGroup`
        All atoms and atom groups within a shell of radius ``radius`` centered
        on ``central_atm_grp``.
    inter_tuples : iterable of \
            (:class:`~luna.interaction.type.InteractionType`, \
             :class:`~luna.mol.groups.AtomGroup`)
        All interactions within a shell of radius ``radius`` centered on
        ``central_atm_grp``. Each tuple contains an
        :class:`~luna.interaction.type.InteractionType` object and one of the
        :class:`~luna.mol.groups.AtomGroup` objects participating to the
        interaction.

        .. note::
            As an interaction involves two participants, it would be expected
            that each interaction produces two tuples. However, by default,
            ShellGenerator sorts atom groups and considers only the first tuple
            that appears, which guarantees that only one of the possible tuples
            is added to avoid information duplication.
    diff_comp_classes : bool
        If True (the default), include differentiation between
        compound classes. That means structural information originated from
        :class:`~luna.mol.groups.AtomGroup` objects belonging to residues,
        nucleotides,  ligands, or water molecules will be considered different
        even if their structural information are the same. This is useful for
        example to differentiate protein-ligand interactions from
        residue-residue ones.
    dtype : data-type
        Use arrays of type ``dtype`` to store information.
        The default value is np.int64.
    seed : int
        A seed to generate shell identifiers through the MurmurHash3 hash
        function. The default value is 0.
    manager : `ShellManager`
        The `ShellManager` object that stores and controls this `Shell` object.
    valid : bool
        If the shell is valid or not. By default, all shells are considered
        valid.
    feature_mapper : dict, optional
        A dict that maps atoms and interactions to unique values.
        If not provided, ``feature_mapper`` will inherit from the default
        mappings ``CHEMICAL_FEATURE_IDS`` and ``INTERACTION_IDS``.

    Attributes
    ----------
    central_atm_grp : :class:`~luna.mol.groups.AtomGroup`
    level : int
    radius : float
    diff_comp_classes : bool
    dtype : data-type
    seed : int
    valid : bool
    feature_mapper : dict
    """

    def __init__(self, central_atm_grp,
                 level, radius, neighborhood=None, inter_tuples=None,
                 diff_comp_classes=True, dtype=np.int64, seed=0, manager=None,
                 valid=True, feature_mapper=None):

        self.central_atm_grp = central_atm_grp
        self.level = level
        self.radius = radius
        self.valid = valid
        self.diff_comp_classes = diff_comp_classes

        if not neighborhood:
            # If inter_tuples was not defined, initialize the neighborhood
            # as an empty list. Otherwise, define the neighborhood as the
            # atom groups interacting with the central atom group.
            neighborhood = ([] if not inter_tuples
                            else [x[1] for x in self._inter_tuples])
        self._neighborhood = set(neighborhood)

        # Always add the central atom in its own neighborhood.
        self._neighborhood.add(central_atm_grp)

        inter_tuples = inter_tuples or []
        self._inter_tuples = set(inter_tuples)

        self._interactions = set([x[0] for x in self._inter_tuples])

        # TODO: FIX IT.
        if not feature_mapper:
            default_dict = {**CHEMICAL_FEATURE_IDS, **INTERACTION_IDS}
            # TODO: Built a class for managing the feature maps.
            # TODO: Attribute a unique id for each new feature.
            # feature_mapper = defaultdict(lambda: -1, default_dict)
            feature_mapper = default_dict
        self.feature_mapper = feature_mapper

        self.dtype = dtype
        self.seed = seed

        self._manager = manager

        # Encode the interactions. It represents the substructures comprised by
        # this shell.
        self._encoded_data = self._encode_interactions()
        # Defined in the hash function
        self._identifier = self.hash_shell()

    @property
    def neighborhood(self):
        """iterable of :class:`~luna.mol.groups.AtomGroup`, read-only: \
                All atoms and atom groups within this shell."""
        return self._neighborhood

    @property
    def interactions(self):
        """iterable of :class:`~luna.interaction.type.InteractionType`, \
                read-only: All interactions within this shell."""
        return self._interactions

    @property
    def inter_tuples(self):
        """iterable of tuple, read-only: Each tuple contains an \
                :class:`~luna.interaction.type.InteractionType` object \
                and one of the :class:`~luna.mol.groups.AtomGroup` objects \
                participating to the interaction."""
        return self._inter_tuples

    @property
    def manager(self):
        """`ShellManager`, read-only: The `ShellManager` object that stores \
        and controls this `Shell` object."""
        return self._manager

    @property
    def identifier(self):
        """int, read-only: This shell identifier, which is generated by \
        hashing its encoded data with a hash function. By default, LUNA \
        uses MurmurHash3 as the hash function."""
        return self._identifier

    @property
    def encoded_data(self):
        """iterable of tuple, read-only: The data encoded in this shell."""
        return self._encoded_data

    @property
    def previous_shell(self):
        """`Shell`, read-only: The previous shell, i.e., a shell centered \
        on the same central :class:`~luna.mol.groups.AtomGroup` object from \
        a previous level. For example, if this shell is in level 5, return \
        a shell from level 4 having the same center."""
        shell = self._manager.get_previous_shell(self.central_atm_grp,
                                                 self.level)
        if shell is None:
            error_msg = ("No previous shell centered in '%s' was found."
                         % self.central_atm_grp)
            logger.exception(error_msg)
            raise ShellCenterNotFound(error_msg)
        return shell

    def is_valid(self):
        """If the shell is valid or not.

        Returns
        -------
         : bool
        """
        return self.valid

    def is_similar(self, shell):
        """ If this shell is similar to ``shell``.

        Two shells are similar if they represent the same substructural
        information.

        Parameters
        ----------
        shell : `Shell`

        Returns
        -------
         : bool
        """

        # The shells' identifiers will be equal when the shells encode the same
        # information and were constructed in the same level.
        if self.level == shell.level:
            if self.identifier == shell.identifier:
                return True
            return False

        # If none of the shells have interactions, check if the shells have
        # equal identifiers.
        if not self.interactions and not shell.interactions:
            # If the identifiers are different, but the shells' central group
            # is the same, it means the shells encode the same information
            # even if their levels are different.
            #
            # OBS: this test is only for fast identification without doing a
            # recursive procedure as the one applied in the ELSE statement.
            if self.central_atm_grp == shell.central_atm_grp:
                return True
            # Although two shells contain different identifiers and their
            # centroids are not the same, they can still be equal if they
            # were obtained in different levels.
            else:
                if self.level > shell.level:
                    return self.previous_shell.is_similar(shell)
                return self.is_similar(shell.previous_shell)

        # If both shells have interactions, check if the interactions are
        # equal.
        elif self.interactions and shell.interactions:
            if self.encoded_data == shell.encoded_data:
                return True
        return False

    def hash_shell(self):
        """Hash this shells' substructural information into a 32-bit integer
        using MurmurHash3.

        Returns
        -------
         : int
            A 32-bit integer representing this shell's substructural
            information.
        """

        if self.level == 0:
            # Get the initial data for the first shell according to the IFP
            # type the user has chosen.
            data = self._initial_shell_data()
        else:
            cent_prev_id = self.previous_shell.identifier

            # Initialization of a new feature vector
            data = [(self.level, cent_prev_id)]

            inter_tuples = []
            for (inter, nb_atm_grp) in self._inter_tuples:
                if isinstance(nb_atm_grp, PseudoAtomGroup):
                    # Pseudo-groups that represents only the atoms involved in
                    # a specific interaction don't generate shells.
                    # Only their parents do, i.e., the whole group generates
                    # shells. So, use it instead.
                    nb_atm_grp = nb_atm_grp.parent_grp

                prev_nb_shell = \
                    self._manager.get_previous_shell(nb_atm_grp, self.level)
                if prev_nb_shell is None:
                    error_msg = ("No previous shell centered in %s was found."
                                 % nb_atm_grp)
                    logger.exception(error_msg)
                    raise ShellCenterNotFound(error_msg)

                # 1st elem: interaction type.
                # 2nd elem: previous identifier of the neighbor atom group;
                inter_tuples.append((self.feature_mapper[inter.type],
                                     prev_nb_shell.identifier))

            # Sort the tuples to avoid dependence on the order in which tuples
            # are added.
            sorted_list = sorted(inter_tuples)

            # Join the interaction information to the feature vector.
            data += sorted_list

        np_array = np.array(data, self.dtype)

        # TODO: Let the user define a hash function
        hashed_shell = mmh3.hash(np_array, self.seed, signed=False)

        return hashed_shell

    def _initial_shell_data(self):
        data = []

        # EIFP uses atomic invariants.
        if self.manager.ifp_type == IFPType.EIFP:
            # Shells use atomic invariants as data. In case of atom groups,
            # the data consists of a list of invariants.
            data = sorted([atm.invariants
                           for atm in self.central_atm_grp.atoms])

        # HIFP uses atomic invariants for atoms and pharmacophore information
        # for atom groups.
        elif self.manager.ifp_type == IFPType.HIFP:

            if len(self.central_atm_grp.atoms) == 1:
                # Shells whose centroid are atoms use invariants as data.
                data = sorted([atm.invariants
                               for atm in self.central_atm_grp.atoms])
            else:
                # Shells whose centroid are atoms' group use pharmacophore
                # as data.
                data = [[self.feature_mapper[cf.format_name()]
                         for cf in self.central_atm_grp.features]]

        # FIFP uses pharmacophore properties for atoms and atoms' group.
        elif self.manager.ifp_type == IFPType.FIFP:

            atm_grp_data = [self.feature_mapper[cf.format_name()]
                            for cf in self.central_atm_grp.features]

            if len(self.central_atm_grp.atoms) == 1:
                features = set()
                # It will loop only through one atom and, therefore, it can be
                # removed without losing information. However, if someday I
                # decide to remove the If, then the code for capturing all atom
                # derived features will be already working.
                for atm in self.central_atm_grp.atoms:
                    for atm_grp in atm.atm_grps:
                        if atm_grp != self.central_atm_grp:
                            features.update(atm_grp.features)

                atm_grp_data += [self.feature_mapper[cf.format_name()]
                                 for cf in features]

            data = [sorted(atm_grp_data)]

        # Include differentiation between compound classes, i.e., groups
        # belonging to Residues, Nucleotides, Ligands, or Waters are
        # treated as being different even when the group would be considered
        # the same.
        if self.diff_comp_classes:
            # `classes` is a list because an atom group can be formed by
            # multiple compounds, which can be from different (e.g., residue
            # and ligand bound covalently) or same (e.g., amide groups formed
            # by backbone residues) classes.
            classes = sorted([CompoundClassIds[c.get_class().upper()].value
                              for c in self.central_atm_grp.compounds])

            np_data = np.array(data, self.dtype)
            np_classes = np.array(classes, self.dtype)

            # Add padding to np_classes
            if np_data.shape[1] > np_classes.shape[0]:
                np_classes = np.pad(np_classes, (0, (np_data.shape[1] - np_classes.shape[0])), 'constant', constant_values=PAD_DEFAULT)
            # Add padding to np_data.
            elif np_data.shape[1] < np_classes.shape[0]:
                np_data = np.pad(np_data, ((0, 0), (0, np_classes.shape[0] - np_data.shape[1])), 'constant', constant_values=PAD_DEFAULT)

            data = np.vstack((np_data, np_classes))

        return data

    def _encode_interactions(self):
        encoded_data = []
        for inter in self.interactions:
            init_src_shell = \
                self._manager.get_previous_shell(inter.src_grp, 1)
            init_trgt_shell = \
                self._manager.get_previous_shell(inter.trgt_grp, 1)
            shell_ids = tuple(sorted([init_src_shell.identifier,
                                      init_trgt_shell.identifier]))
            encoded_data.append((shell_ids, self.feature_mapper[inter.type]))
        return sorted(encoded_data)

    def __repr__(self):
        return ("<Shell: level=%d, radius=%f, center=%s, interactions=%d>"
                % (self.level, self.radius,
                   self.central_atm_grp, len(self.interactions)))


class ShellGenerator:

    """Generate shells, the base information of LUNA fingerprints.

    Parameters
    ----------
    num_levels : int
        The maximum number of iterations for fingerprint generation.
    radius_step : float
        The multiplier used to increase shell size at each iteration.
        At iteration 0, shell radius is 0 * ``radius_step``, at iteration 1,
        radius is 1 * ``radius_step``, etc.
    fp_length : int
        The fingerprint length (total number of bits).
        The default value is :math:`2^{32}`.
    ifp_type : :class:`~luna.interaction.fp.type.IFPType`
        The fingerprint type (EIFP, FIFP, or HIFP). The default value is EIFP.
    diff_comp_classes : bool
        If True (the default), include differentiation between compound
        classes. That means structural information originated from
        :class:`~luna.mol.groups.AtomGroup` objects belonging to residues,
        nucleotides,  ligands, or water molecules will be considered different
        even if their structural information are the same. This is useful for
        example to differentiate protein-ligand interactions from
        residue-residue ones.
    dtype : data-type
        Use arrays of type ``dtype`` to store information.
        The default value is np.int64.
    seed : int
        A seed to generate shell identifiers through the MurmurHash3
        hash function. The default value is 0.
    bucket_size : int
        Bucket size of KD tree.
        You can play around with this to optimize speed if you feel like it.
        The default value is 10.

    Attributes
    ----------
    num_levels : int
    radius_step : float
    fp_length : int
    ifp_type : :class:`~luna.interaction.fp.type.IFPType`
    diff_comp_classes : bool
    dtype : data-type
    seed : int
    bucket_size : int

    Examples
    --------

    In the below example, we will assume a LUNA project object named
    ``proj_obj`` already exists.

    First, let's define a `ShellGenerator` object that will create shells over
    2 iterations (levels). At each iteration, the shell radius will be
    increased by 3 and substructural information will be encoded following EIFP
    definition.

    >>> from luna.interaction.fp.shell import ShellGenerator
    >>> from luna.interaction.fp.type import IFPType
    >>> num_levels, radius_step = 2, 3
    >>> sg = ShellGenerator(num_levels, radius_step, ifp_type=IFPType.EIFP)

    After defining the generator, we can create shells by calling
    :meth:`create_shells`, which expects an
    :class:`~luna.mol.groups.AtomGroupsManager` object. In this example, we
    will the first :class:`~luna.mol.groups.AtomGroupsManager` object from an
    existing LUNA project (``proj_obj``).

    >>> atm_grps_mngr = list(proj_obj.atm_grps_mngrs)[0]
    >>> sm = sg.create_shells(atm_grps_mngr)
    >>> print(sm.num_shells)
    528

    Now, with shells stored in the `ShellManager` object you can, for instance:

        * Generate fingerprints:

            >>> fp = sm.to_fingerprint(fold_to_length=1024)
            >>> print(fp.indices)
            [   2   19   22   23   34   37   39   45   54   67   71   75   83   84
               93  109  138  140  157  162  181  182  186  187  191  194  206  209
              211  237  246  251  263  271  281  296  304  315  323  358  370  374
              388  392  399  400  419  439  476  481  487  509  519  527  532  578
              587  592  604  605  629  635  645  661  668  698  711  713  732  736
              740  753  764  795  813  815  820  824  825  831  836  855  873  882
              911  926  967  975  976  984  990  996 1020]

        * Visualize substructural information in Pymol\:

            >>> from luna.interaction.fp.view import ShellViewer
            >>> shell_tuples = [(atm_grps_mngr.entry,
            ...                  sm.unique_shells,
            ...                  proj_obj.pdb_path)]
            >>> sv = ShellViewer()
            >>> sv.new_session(shell_tuples, "example.pse")
    """

    def __init__(self, num_levels, radius_step,
                 fp_length=DEFAULT_FP_LENGTH, ifp_type=IFPType.EIFP,
                 diff_comp_classes=True, dtype=np.int64, seed=0,
                 bucket_size=10):

        self.num_levels = num_levels
        self.radius_step = radius_step
        self.fp_length = fp_length
        self.ifp_type = ifp_type
        self.diff_comp_classes = diff_comp_classes

        self.seed = seed
        self.dtype = dtype
        self.bucket_size = bucket_size

    def create_shells(self, atm_grps_mngr):
        """Perceive substructural information from
        :class:`~luna.mol.groups.AtomGroup` objects and their interactions,
        and represent such information as shells.

        Parameters
        ----------
        atm_grps_mngr : :class:`~luna.mol.groups.AtomGroupsManager`
            Container of :class:`~luna.mol.groups.AtomGroup` objects and their
            interactions.

        Returns
        -------
         : `ShellManager`

        Raises
        ------
        ShellCenterNotFound
            If it fails to recover a shell having a given center.
        """

        sm = ShellManager(self.num_levels,
                          self.radius_step,
                          self.fp_length,
                          self.ifp_type)

        all_interactions = atm_grps_mngr.get_all_interactions()

        neighborhood = set(atm_grps_mngr.atm_grps)

        # Create pseudo-groups for hydrophobic interactions.
        pseudo_grps = set(self._generate_pseudo_grps(atm_grps_mngr))

        # Create a neighborhood scope for the searches
        search_scope = neighborhood | pseudo_grps
        nbs = AtomGroupNeighborhood(search_scope, self.bucket_size)

        # Map an atom to all of its pseudo-groups.
        pseudo_grps_mapping = defaultdict(set)
        for pseudo_grp in pseudo_grps:
            atoms = tuple(sorted(pseudo_grp.atoms))
            for atm in atoms:
                # Add an atom to the mapping
                pseudo_grps_mapping[(atm,)].add(pseudo_grp)
            # Add a list of atoms
            pseudo_grps_mapping[atoms].add(pseudo_grp)

        # It controls which atom groups already converged.
        skip_atm_grps = set()

        # Sort the atom groups to avoid order dependence.
        sorted_neighborhood = sorted(neighborhood)

        level = -1
        for level in range(self.num_levels):
            radius = self.radius_step * level

            for atm_grp in sorted_neighborhood:

                # Ignore centroids that already reached the limit of possible
                # substructures.
                if atm_grp in skip_atm_grps:
                    continue

                # It stores all possible expansions each group can do.
                # Each expansion is a derived group. Initially, the list
                # contains only the groups derived from the central atom
                # group. But, later it may also contain derived groups
                # from interacting partner groups.
                all_derived_atm_grps = \
                    self._get_derived_grps(atm_grp, pseudo_grps_mapping)

                # In level 0, the number of unique derived groups is 0 as the
                # shell initially only contains information of the centroid.
                unique_derived_atm_grps = []

                shell = None

                if radius > 0:
                    prev_shell = sm.get_previous_shell(atm_grp, level)
                    if not prev_shell:
                        error_msg = ("No previous shell centered in %s "
                                     "was found." % atm_grp)
                        logger.exception(error_msg)
                        raise ShellCenterNotFound(error_msg)

                    prev_atm_grps = prev_shell.neighborhood
                    prev_interactions = prev_shell.interactions

                    nb_atm_grps = set(nbs.search(atm_grp.centroid, radius))

                    inter_tuples = set()
                    interactions_to_add = set()

                    # For each atom group from the previous shell.
                    for prev_atm_grp in sorted(prev_atm_grps):
                        # Include the partner's derived groups to the set of
                        # all derived groups.
                        all_derived_atm_grps.update(self._get_derived_grps(prev_atm_grp, pseudo_grps_mapping))

                        for inter in prev_atm_grp.interactions:
                            # Only PseudoAtomGroup objects should deal with
                            # hydrophobic interactions.
                            if isinstance(prev_atm_grp, PseudoAtomGroup) is False and inter.type == "Hydrophobic":
                                continue

                            partner_grp = self._recover_partner_grp(prev_atm_grp, inter, pseudo_grps_mapping)

                            if partner_grp is not None:
                                if inter not in interactions_to_add and partner_grp in nb_atm_grps:
                                    new_tuple = (inter, partner_grp)

                                    # Ignore interactions that already exists
                                    # in the previous shell to avoid
                                    # duplications in the list of interactions.
                                    # For example, without this control, an
                                    # interaction I between atom A1 and A2
                                    # would appear twice in the list: (I, A1)
                                    # and (I, A2). Thus, it keeps only the
                                    # first interaction that appears while
                                    # increasing the shell.
                                    if inter in prev_interactions and new_tuple not in prev_shell.inter_tuples:
                                        continue

                                    inter_tuples.add(new_tuple)
                                    interactions_to_add.add(inter)

                    # Get valid derived groups, which are those ones inside the
                    # current sphere.
                    valid_derived_atm_grps = \
                        set([ag for ag in all_derived_atm_grps
                             if ag in nb_atm_grps])

                    unique_derived_atm_grps = \
                        valid_derived_atm_grps - prev_atm_grps

                    # It adds a new shell only if there are new interactions
                    # and derived atom groups inside the shell.
                    shell_nb = set([x[1] for x in inter_tuples]) | valid_derived_atm_grps
                    shell_nb.add(atm_grp)

                    shell = Shell(atm_grp, level, radius,
                                  neighborhood=shell_nb,
                                  inter_tuples=inter_tuples,
                                  manager=sm,
                                  diff_comp_classes=self.diff_comp_classes,
                                  seed=self.seed, dtype=self.dtype)
                else:
                    shell = Shell(atm_grp, level, radius, manager=sm,
                                  diff_comp_classes=self.diff_comp_classes,
                                  seed=self.seed, dtype=self.dtype)

                if shell:
                    sm.add_shell(shell)
                    last_shell = shell
                else:
                    last_shell = sm.get_last_shell(atm_grp)

                # Evaluate if the limit of possible substructures for the
                # current centroid (atom group) was reached.
                if last_shell:
                    # The limit will be reached when the last shell already
                    # contains all interactions established by the atom
                    # groups inside the shell. In this case, expanding the
                    # radius will not result in any new shell because a shell
                    # is only created when the atoms inside the last shell
                    # establish interactions with the atom groups found after
                    # increasing the radius.
                    all_nb_interactions = set(chain.from_iterable([g.interactions for g in last_shell.neighborhood]))
                    # It considers only interactions whose atom groups exist in
                    # the neighborhood.
                    valid_interactions = set([i for i in all_nb_interactions if i.src_grp in neighborhood and i.trgt_grp in neighborhood])

                    # It identifies if the convergence for the interactions
                    # was reached.
                    interactions_converged = valid_interactions == last_shell.interactions

                    # It identifies if the convergence for the expansion of
                    # atom groups was reached, which happens when all derived
                    # groups were already included in the centroid
                    # neighborhood. However, it may happen that this
                    # requirement was fulfilled by the time new unique derived
                    # groups were included in the centroid neighborhood.
                    # Thus, the second test provides a chance to all of these
                    # recently discovered groups to expand.
                    grp_expansions_converged = all_derived_atm_grps == last_shell.neighborhood and len(unique_derived_atm_grps) == 0

                    # The local convergence is reached when no atom group
                    # inside the current sphere can expand or all of its
                    # interactions were already included to the sphere.
                    local_convergence = \
                        interactions_converged and grp_expansions_converged
                    # The global convergence occurs when all interactions in
                    # the binding site were already included in the sphere.
                    global_convergence = \
                        all_interactions == last_shell.interactions

                    # If the limit was reached for this centroid, in the next
                    # level it can be ignored.
                    if local_convergence or global_convergence:
                        skip_atm_grps.add(atm_grp)

            # If all atom groups reached the limit of possible substructures,
            # just leave the loop.
            if len(skip_atm_grps) == len(neighborhood):
                logger.debug("The list of shells cannot be expanded anymore. "
                             "The maximum number of substructures were "
                             "reached.")
                break

        logger.debug("Shells creation finished.")
        logger.debug("The last level executed was: %d." % level)
        logger.debug("The number of levels defined was: %d." % self.num_levels)
        logger.debug("Total number of shells created: %d" % sm.num_shells)
        logger.debug("Total number of unique shells created: %d"
                     % sm.num_unique_shells)

        return sm

    def _generate_pseudo_grps(self, atm_grps_mngr):
        pseudo_grps = {}
        mapped_interactions = set()

        for hydrop_grp in atm_grps_mngr.filter_by_types(["Hydrophobe"]):
            for inter in hydrop_grp.interactions:

                # Ignore non-hydrophobic interactions or interactions already
                # mapped.
                if inter.type != "Hydrophobic" or inter in mapped_interactions:
                    continue

                src_tuple = (inter.src_grp, tuple(sorted(inter.src_interacting_atms)))
                trgt_tuple = (inter.trgt_grp, tuple(sorted(inter.trgt_interacting_atms)))

                for atm_grp, atms in [src_tuple, trgt_tuple]:
                    # It will get a pseudo-group already created or create
                    # a new one.
                    feats = [ChemicalFeature("Hydrophobe")]
                    pseudo_grp = \
                        pseudo_grps.get(atms,
                                        PseudoAtomGroup(atm_grp, atms, feats))

                    pseudo_grp.add_interactions([inter])

                    pseudo_grps[atms] = pseudo_grp

                mapped_interactions.add(inter)

        return pseudo_grps.values()

    def _get_derived_grps(self, atm_grp, pseudo_grps_mapping):
        # Get derived groups for the informed atom group.
        derived_atm_grps = set([ag
                                for a in atm_grp.atoms
                                for ag in a.atm_grps])

        # Get derived pseudo-groups for the informed atom group.
        derived_atm_grps.update(set([pseudo_grp
                                     for a in atm_grp.atoms
                                     for pseudo_grp in pseudo_grps_mapping.get((a,), [])]))

        return derived_atm_grps

    def _recover_partner_grp(self, atm_grp, interaction, pseudo_grps_mapping):
        if isinstance(atm_grp, PseudoAtomGroup):
            atoms = sorted(atm_grp.atoms)

            src_atms = sorted(interaction.src_interacting_atms)
            trgt_atms = sorted(interaction.trgt_interacting_atms)

            if atoms == src_atms:
                partner_atms = trgt_atms
            elif atoms == trgt_atms:
                partner_atms = src_atms
            else:
                return None

            partner_grp = None
            # Recover the partner groups based on the interacting atoms.
            # Then, it will check which returned pseudo-group corresponds to
            # the interacting atoms, i.e., the one that contains exactly the
            # same atoms. This verification is mainly necessary for
            # pseudo-groups composed only by one atom as such groups tend
            # to belong to more than one pseudo-group.
            for pseudo_grp in pseudo_grps_mapping.get(tuple(partner_atms), []):
                if sorted(pseudo_grp.atoms) == partner_atms:
                    partner_grp = pseudo_grp
                    break

            return partner_grp
        else:
            return interaction.get_partner(atm_grp)



from itertools import combinations
from openbabel import openbabel as ob

from Bio.PDB.PDBIO import Select

import logging

logger = logging.getLogger()


def is_covalently_bound(atm1, atm2):
    """Verifies if atoms ``atm1`` and ``atm2`` are covalently bound."""

    # Distance atom-atom
    dist = atm1 - atm2
    # Covalent radius.
    #   Note that we call title() to format atoms' symbol as in Open Babel.
    #   E.g.: ZN becomes Zn.
    cov1 = ob.GetCovalentRad(ob.GetAtomicNum(atm1.element.title()))
    cov2 = ob.GetCovalentRad(ob.GetAtomicNum(atm2.element.title()))

    # OpenBabel thresholds.
    if 0.4 <= dist <= cov1 + cov2 + 0.45:
        return True
    return False


def get_residue_cov_bonds(residue, select=Select()):
    """Get covalently bound atoms from residues or other molecules.

    Parameters
    ----------
    residue : :class:`~luna.pdb.core.residue.Residue`
        The residue or other molecule from which covalently
        bound atoms will be recovered.
    select : :class:`Bio.PDB.PDBIO.Select`
        Decides which atoms will be consired.
        By default, all atoms are accepted.

    Returns
    -------
    list of tuple(:class:`~luna.pdb.core.atom.Atom`, \
            :class:`~luna.pdb.core.atom.Atom`)
        List of pairs of covalently bound atoms.
    """

    # Get valid atoms according to the provided selection function.
    trgt_res_atms = {atm.name: atm for atm in residue.get_atoms()
                     if select.accept_atom(atm)}

    cov_bonds = []
    pairs = combinations(trgt_res_atms.values(), 2)
    for atm1, atm2 in pairs:
        if is_covalently_bound(atm1, atm2):
            cov_bonds.append((atm1, atm2))

    return cov_bonds



from subprocess import Popen, PIPE
import tempfile
import re

from luna.wrappers.chemaxon.util import check_license
from luna.wrappers.base import MolWrapper
from luna.util.exceptions import ProcessingFailed
from luna.util.file import new_unique_filename

import logging
logger = logging.getLogger()


REGEX_MOL_FILE = re.compile(r' (V2000|V3000)$')


class TautomerGenerator:

    def __init__(self, ph=None, path_length=4,
                 protect_aromaticity=True,
                 protect_charge=True,
                 exclude_antiaromatic=True,
                 protect_double_bond_stereo=False,
                 protect_all_tetra_centers=False,
                 protect_labeled_tetra_centers=False,
                 protect_ester_groups=True,
                 min_percentage=20,
                 only_non_zeros=True,
                 mol_obj_type="rdkit",
                 skip_errors=False,
                 cxcalc="cxcalc"):

        self.ph = ph
        self.path_length = path_length
        self.protect_aromaticity = protect_aromaticity
        self.protect_charge = protect_charge
        self.exclude_antiaromatic = exclude_antiaromatic
        self.protect_double_bond_stereo = protect_double_bond_stereo
        self.protect_all_tetra_centers = protect_all_tetra_centers
        self.protect_labeled_tetra_centers = protect_labeled_tetra_centers
        self.protect_ester_groups = protect_ester_groups
        self.min_percentage = min_percentage
        self.only_non_zeros = only_non_zeros
        self.mol_obj_type = mol_obj_type
        self.skip_errors = skip_errors
        self.cxcalc = cxcalc

        self.percentages = []

    def _prep_opts(self):

        opt_list = []

        if self.ph is not None:
            opt_list += ["-H", str(self.ph)]

        # Path length
        opt_list += ["-l", str(self.path_length)]

        # Protect aromaticity
        opt_list += ["-a", str(self.protect_aromaticity)]

        # Protect charge
        opt_list += ["-C", str(self.protect_charge)]

        # Exclude antiaromatic compounds
        opt_list += ["-e", str(self.exclude_antiaromatic)]

        # Protect double bond stereo
        opt_list += ["-P", str(self.protect_double_bond_stereo)]

        # Protect all tetrahedral stereo
        opt_list += ["-T", str(self.protect_all_tetra_centers)]

        # Protect labeled tetrahedral stereo centers
        opt_list += ["-L", str(self.protect_labeled_tetra_centers)]

        # Protect ester groups
        opt_list += ["-E", str(self.protect_ester_groups)]

        return opt_list

    def _parse(self):
        self.tautomers = []
        self.percentages = []

        def apply_mol_format(lines):
            # Check if the MOL file contains a valid header comprising
            # a Title line, Program/file timestamp line, and a Comment line.
            # Thus, the Counts line must be in line 4 (index 3).
            # If it's not, add blank lines to match the format.
            # This amendment is necessary otherwise RDKit will crash.
            if not REGEX_MOL_FILE.search(lines[3]):
                logger.info("While parsing the file '%s', we found a "
                            "molecule starting at line #%d that does "
                            "not contain a valid header. "
                            "We will add empty lines to its header to "
                            "match the MOL file format."
                            % (self.output_file, mol_starts_at))

                counts_line_pos = None
                for i, line in enumerate(lines):
                    if REGEX_MOL_FILE.search(line):
                        counts_line_pos = i
                        break

                missing_lines = ["\n"] * (3 - counts_line_pos)
                lines = missing_lines + lines

            if lines[-1].strip() != "M  END":
                logger.info("While parsing the file '%s', we found a "
                            "molecule starting at line #%d that does not "
                            "contain the END line. We will add it to the "
                            "block end to match the MOL file format."
                            % (self.output_file, mol_starts_at))
                lines.append("M  END\n")
            return lines

        with open(self.output_file) as IN:
            n_mols = 0
            mol = []
            line_count = 0
            ignore_lines = False
            for line in IN:
                line_count += 1

                # Ignore new lines before a molecule block.
                if len(mol) == 0 and line == "\n":
                    continue

                # Save the line in which the new molecule starts.
                if len(mol) == 0:
                    mol_starts_at = line_count

                if line.startswith("M  END"):
                    mol.append(line)
                    # Fix the MOL block in cases where the header or end
                    # lines do not match the MOL file format.
                    mol = apply_mol_format(mol)

                    # After finding a molecule block, ignore any following
                    # lines until it finds a line "$$$$".
                    ignore_lines = True

                elif line.startswith(">  <TAUTOMER_DISTRIBUTION>"):
                    line = IN.readline()
                    line_count += 1
                    # readline() returns empty strings when EOF is reached.
                    if not line:
                        raise StopIteration

                    perc = float(line.strip())

                    # If the tautomer distribution is less than
                    # the minimum percentage or if it is zero,
                    # ignore the molecule.
                    ignore_mol = False
                    if ((self.min_percentage is not None
                            and perc < self.min_percentage)
                            or (self.only_non_zeros and perc == 0)):
                        ignore_mol = True

                    if not ignore_mol:
                        mol_obj = None
                        try:
                            # Create a new MolWrapper object
                            mol_obj = \
                                MolWrapper.from_mol_block("".join(mol), "sdf",
                                                          self.mol_obj_type)
                        except Exception:
                            if not self.skip_errors:
                                raise

                        if mol_obj is not None:
                            mol_id = "Tautomer %d" % (n_mols + 1)
                            mol_obj.set_name(mol_id)

                            self.tautomers.append(mol_obj)
                            self.percentages.append((mol_id, perc))

                    # Restart the molecule block.
                    mol = []

                elif line.startswith("$$$$") is False:
                    # Ignore lines util it finds a line "$$$$".
                    if ignore_lines:
                        continue
                    mol.append(line)
                else:
                    ignore_lines = False
                    n_mols += 1

    def _run_cxcalc(self, mol_input, output_file=None):
        opt_list = self._prep_opts()

        self.mol_input = mol_input

        if output_file is None:
            output_file = new_unique_filename(tempfile.gettempdir()) + ".sdf"
        self.output_file = output_file
        output = ["-o", output_file]

        plugin = "dominanttautomerdistribution"
        args = [self.cxcalc] + [mol_input] + output + [plugin] + opt_list

        p = Popen(args, stdout=PIPE, stderr=PIPE)
        try:
            stdout, stderr = p.communicate()
        except Exception:
            p.kill()
            raise

        error_lines = stderr.decode().strip().split("\n")

        if len(error_lines) == 1 and error_lines[0] == "":
            logger.debug("Tautomers for input '%s' created with success."
                         % mol_input)
        else:
            logger.error(stderr.decode())
            raise ProcessingFailed("cxcalc could not generate tautomers. "
                                   "Check the logs for more information.")

        self._parse()

    def generate(self, mol_input, output_file=None):
        self.tautomers = []

        check_license()

        self._run_cxcalc(mol_input, output_file)

        return self.tautomers



from pathlib import Path
from subprocess import Popen, PIPE
from os.path import exists

from luna.util.exceptions import IllegalArgumentError, ProcessingFailed
from luna.wrappers.chimera.add_charges import CHARGE_METHODS
from luna.util.file import is_file_valid


import logging
logger = logging.getLogger()


def add_charges(input_file, output_file, total_charge,
                charge_method="am1-bcc"):

    logger.debug("Charges will be added to the molecule at '%s'." % input_file)

    if not is_file_valid(input_file):
        raise OSError("File '%s' does not exist.")

    if not isinstance(output_file, str):
        raise IllegalArgumentError("'output_file' must be a string.")

    if charge_method not in CHARGE_METHODS:
        raise IllegalArgumentError("The charge method '%s' is not "
                                   "accepted by Chimera." % charge_method)

    script = str(Path(__file__).parent) + "/add_charges.py"
    command = (f"chimera --nogui --nostatus --script '{script} "
               f"-i \"{input_file}\" -o \"{output_file}\" "
               f"-nc {total_charge} -c {charge_method}'")

    p = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
    try:
        stdout, stderr = p.communicate()
    except Exception:
        p.kill()
        raise

    if exists(output_file):
        logger.debug("File '%s' created with success." % output_file)
    else:
        logger.error(stderr.decode())
        raise ProcessingFailed("Chimera could not add charges "
                               "to the provided molecule.")



from Bio.PDB.Model import Model as BioModel

from luna.pdb.core.entity import Entity


class Model(BioModel, Entity):
    """
    Custom Model subclass with SMCRA traversal support.

    Parameters
    ----------
    id : int
        Model index.
    serial_num : int, optional
        Optional serial number for PDB output.
    """

    def __init__(self, id, serial_num=None):
        super().__init__(id, serial_num)

    def __repr__(self):
        return f"<Model (LUNA subclass) id={self.id}>"

    def __lt__(self, other):
        """Allow sorting of models by ID."""
        return self.id < other.id

    def get_atoms(self):
        """Return atoms."""
        for r in self.get_residues():
            for a in r.get_unpacked_list():
                yield a



import os

from luna.util.exceptions import LicenseNotFoundError


def check_license():

    if ("CHEMAXON_LICENSE_URL" not in os.environ
            and "CHEMAXON_LICENSE_SERVER_KEY" not in os.environ):
        raise LicenseNotFoundError("The Chemaxon license is not installed.")



import tempfile
import logging

from Bio.PDB import Selection
from Bio.PDB.PDBIO import Select

from openbabel.pybel import readfile
from openbabel.pybel import Molecule as PybelWrapper

from rdkit.Chem import (MolFromMolBlock, SanitizeFlags,
                        SanitizeMol, MolToMolFile)


from luna.pdb.io.helpers import save_to_file
from luna.pdb.io.selector import AtomSelector
from luna.mol.validator import MolValidator
from luna.mol.standardiser import Standardizer
from luna.wrappers.base import MolWrapper
from luna.wrappers.rdkit import read_mol_from_file
from luna.wrappers.obabel import convert_molecule
from luna.util.file import new_unique_filename, remove_files
from luna.util.exceptions import (MoleculeSizeError,
                                  MoleculeObjectError)
from luna.util.default_values import OPENBABEL


logger = logging.getLogger(__name__)


def biopython_entity_to_mol(entity,
                            select=Select(),
                            amend_mol=True,
                            template=None,
                            add_h=False,
                            ph=None,
                            metals_coord=None,
                            wrapped=True,
                            openbabel=OPENBABEL,
                            tmp_path=None,
                            keep_tmp_files=False):
    """Convert an object :class:`~luna.pdb.core.entity.Entity` to a
    molecular object (:class:`~luna.wrappers.base.MolWrapper`,
    :class:`rdkit.Chem.rdchem.Mol`, or :class:`openbabel.pybel.Molecule`).

    Parameters
    ----------
    entity : :class:`~luna.pdb.core.entity.Entity`
        The entity to be converted.
    select : :class:`Bio.PDB.PDBIO.Select`
        Decide which atoms will be considered.
        By default, all atoms are accepted.
    amend_mol : bool
        If True, validate and standardize molecules.
        The validation is employed with
        :class:`~luna.mol.validator.MolValidator`.
        Currently, only residues are standardized.
    template : :class:`luna.mol.template.Template`, optional
        Fix the converted molecule's bond order based on the bond orders
        in a template molecule. The ``template`` should implement
        assign_bond_order().
    add_h : bool
        If True, add hydrogen to the converted molecule.
    ph : float, optional
        Add hydrogens considering pH ``ph``.
    wrapped : bool
        If True, wrap the molecular object with
        :class:`~luna.wrappers.base.MolWrapper`.
    openbabel : str
        Pathname to Open Babel.
    tmp_path : str
        A temporary directory to where temporary files will be saved.
        If not provided, the system's default temporary directory will
        be used instead.
    keep_tmp_files : bool
        If True, keep all temporary files. Otherwise, removes them in the end.

    Returns
    -------
    mol_obj : :class:`~luna.wrappers.base.MolWrapper`, \
                :class:`rdkit.Chem.rdchem.Mol`, \
                or :class:`openbabel.pybel.Molecule`
        The converted molecule.
    ignored_atoms : list
        List of ignored atoms. Currently, ignored atoms may
        contain only metals.
    """

    tmp_path = tmp_path or tempfile.gettempdir()

    logger.debug("It will try to create a new MOL object from the provided "
                 "entity.")
    logger.debug("Temporary files will be saved at '%s'." % tmp_path)

    # First it saves the selection into a PDB file and then it converts the
    # file to .mol. I had to do it because the OpenBabel 2.4.1 had a
    # problem with some molecules containing aromatic rings. In such cases,
    # the aromatic ring was being wrongly perceived and some atoms received
    # more double bonds than it was expected. The version 2.3.2 works better.
    # Therefore, I recomend using Open Babel 2.3.3 instead.
    filename = new_unique_filename(tmp_path)
    pdb_file = '%s_pdb-file.pdb' % filename

    logger.debug("First: it will try to create a new PDB file (%s) "
                 "from the provided entity." % pdb_file)

    # Apparently, Open Babel creates a bug when it tries to parse a file with
    # CONECTS containing serial numbers with more than 4 digits.
    # E.g.: 1OZH:A:HE3:1406, line CONECT162811627916282.
    # By setting preserve_atom_numbering to False, it solves the problem.
    save_to_file(entity,
                 pdb_file,
                 select,
                 preserve_atom_numbering=False,
                 sort=True)

    ini_input_file = pdb_file
    if template is not None:
        if entity.level == "R" and entity.is_hetatm():
            # Note that the template molecule should have no explicit hydrogens
            # else the algorithm will fail.
            rdmol = read_mol_from_file(pdb_file, mol_format="pdb",
                                       removeHs=True)
            new_rdmol = template.assign_bond_order(rdmol, entity.resname)

            ini_input_file = '%s_tmp-mol-file.mol' % filename
            MolToMolFile(new_rdmol, ini_input_file)

            if not keep_tmp_files:
                remove_files([pdb_file])
        else:
            logger.warning("It cannot apply a template on the provided entity "
                           "because it should be a single compound "
                           "(Residue class).")

    # Convert the PDB file to Mol file with the proper protonation
    # and hydrogen addition if required.
    mol_file = '%s_mol-file.mol' % filename
    ob_opt = {"error-level": 5}
    logger.debug("Next: it will try to convert the PDB file to "
                 " .mol using Open Babel.")
    if add_h:
        logger.debug("Hydrogens will be added to the molecule.")
        if ph is not None:
            ob_opt["p"] = ph
        else:
            ob_opt["h"] = ""
    convert_molecule(ini_input_file, output_file=mol_file,
                     opts=ob_opt, openbabel=openbabel)

    # Currently, ignored atoms are only metals.
    ignored_atoms = []

    mol_obj = None
    if amend_mol:
        logger.debug("Now: a validation will be performed and "
                     "it will try to fix some errors.")

        try:
            mol_obj = next(readfile("mol", mol_file))
        except Exception:
            error_msg = ("An error occurred while parsing the file '%s' with "
                         "Open Babel and the molecule object could not be "
                         "created. Check the logs for more information."
                         % mol_file)
            raise MoleculeObjectError(error_msg)

        # Standardize a specific set of atoms if provided,
        # otherwise use all atoms from entity.
        if isinstance(select, AtomSelector):
            target_atoms = select.entries
        else:
            target_atoms = Selection.unfold_entities(entity, 'A')

        # If the add_h property is set to False, the code will not remove
        # any existing hydrogens from the PDB structure.
        # In these situations, the list of atoms may contain hydrogens.
        # But, we do not need to attribute properties to hydrogens.
        # We just need them to correctly set properties to heavy atoms.
        # So let's just ignore them.
        target_atoms = [atm for atm in target_atoms
                        if select.accept_atom(atm) and atm.element != "H"]

        mol_obj = MolWrapper(mol_obj)
        if mol_obj.get_num_heavy_atoms() != len(target_atoms):
            error_msg = ("The number of heavy atoms in the PDB selection "
                         "and in the MOL file are different.")
            raise MoleculeSizeError(error_msg)

        # Ignore hydrogen atoms.
        atm_obj_list = [atm for atm in mol_obj.get_atoms()
                        if atm.get_atomic_num() != 1]

        # Pairs of AtomWrapper and Atom (Biopython) objects
        # to standardize.
        atom_pairs = []
        pdb_mapping = {}
        for i, atm_obj in enumerate(atm_obj_list):
            atom_pairs.append((atm_obj, target_atoms[i]))

            pdb_mapping[atm_obj.get_idx()] = target_atoms[i]

        updated_metals_coord = {}
        # If there is something to be standardised.
        if atom_pairs:
            rs = Standardizer()
            mol_obj.unwrap().DeleteHydrogens()
            rs.standardize(atom_pairs, metals_coord=metals_coord)

            updated_metals_coord = rs.metals_coord

            # After standardizing residues, we need to recreate the Mol
            # file, otherwise implicit hydrogens will not be included
            # in the MOL object and, therefore, their coordinates could
            # not be accessed. If you try to generate coordinates directly
            # from the object, hydrogens will be incorrectly placed.
            mol_obj = PybelWrapper(mol_obj.unwrap())
            new_mol_file = '%s_tmp-mol-file.mol' % filename
            mol_obj.write("mol", new_mol_file, overwrite=True)

            # Overwrite mol_file by converting the new molecular file using
            # the user specified parameters. Note that right now it will add
            # explicit hydrogens to the molecules according to the provided pH.
            convert_molecule(new_mol_file, output_file=mol_file, 
                             opts=ob_opt, openbabel=OPENBABEL)

            # Let's finally read the correct and standardized molecular file.
            try:
                mol_obj = next(readfile("mol", mol_file))
            except Exception:
                error_msg = ("An error occurred while parsing the file "
                             "'%s' with Open Babel and the molecule object"
                             "could not be created. Check the logs for "
                             "more information." % mol_file)
                raise MoleculeObjectError(error_msg)

            # Remove temporary files.
            if not keep_tmp_files:
                remove_files([new_mol_file])

        mv = MolValidator(metals_coord=updated_metals_coord)
        is_valid = mv.validate_mol(mol_obj, pdb_mapping)
        logger.debug('Validation finished!!!')

        if not is_valid:
            logger.warning("The molecular file '%s' contain invalid atoms. "
                           "Check the logs for more information." % mol_file)

        # Validate molecule using RDKit sanitization methods.
        try:
            aux_mol = (PybelWrapper(mol_obj.unwrap())
                       if isinstance(mol_obj, MolWrapper)
                       else mol_obj)

            # The sanitization is set off. We will apply it in the next step.
            aux_mol = MolFromMolBlock(aux_mol.write('mol'),
                                      sanitize=False, removeHs=False)
            # Sanitize molecule is applied now, so we will be able to catch
            # the exceptions raised by RDKit, otherwise it would
            # not be possible.
            SanitizeMol(aux_mol, SanitizeFlags.SANITIZE_ALL)
        except Exception:
            error_msg = ("An error occurred while parsing the molecular block "
                         "with RDKit. The block was generated by Open Babel "
                         "from the file '%s'. Check the logs for more "
                         "information." % mol_file)
            raise MoleculeObjectError(error_msg)
    else:
        try:
            # Create a new Mol object.
            mol_obj = next(readfile("mol", mol_file))
        except Exception:
            error_msg = ("An error occurred while parsing the file '%s' and "
                         "the molecule object could not be created. "
                         "Check the logs for more information."
                         % mol_file)
            raise MoleculeObjectError(error_msg)

    # Remove temporary files.
    if not keep_tmp_files:
        remove_files([ini_input_file, mol_file])

    if wrapped:
        mol_obj = MolWrapper(mol_obj)
    elif isinstance(mol_obj, MolWrapper):
        mol_obj = mol_obj.unwrap()

    return mol_obj, ignored_atoms


import logging

from Bio.PDB.Model import Model
from Bio.PDB.Chain import Chain

from luna.pdb.io.helpers import save_to_file
from luna.pdb.io.selector import ChainSelector, ResidueSelector


logger = logging.getLogger()


class Extractor():
    """
    Extract chains or residues from a model or chain and save them to a PDB file.

    Parameters
    ----------
    entity : :class:`~luna.pdb.core.model.Model` or :class:`~luna.pdb.core.chain.Chain`
        A model or chain from which entities will be extracted.
    """

    def __init__(self, entity):
        self.entity = entity

    @property
    def entity(self):
        return self._entity

    @entity.setter
    def entity(self, entity):
        if not isinstance(entity, (Model, Chain)):
            raise TypeError("Only objects of type %s or %s are accepted." %
                            (Model.__module__, Chain.__module__))
        self._entity = entity

    def extract_chains(self, chains, output_file):
        """
        Extract specific chains from a model and save to a file.

        Parameters
        ----------
        chains : iterable of str
            Chain IDs to extract.
        output_file : str
            Filepath to write the selected chains.
        """
        
        if self.entity.level != "M":
            logger.warning("Entity must be a Model object.")
            return
        
        chain_sel = set()
        for chain in chains:
            if chain in self.entity.child_dict:
                chain_sel.add(self.entity[chain])
            else:
                logger.warning("Chain %s does not exist." % chain)

        if len(chain_sel) > 0:
            save_to_file(self.entity,
                         output_file,
                         ChainSelector(chain_sel))
        else:
            logger.warning("No valid chain to extract.")

    def extract_residues(self, residues, output_file):
        """
        Extract specific residues from a chain and save to a file.

        Parameters
        ----------
        residues : iterable of tuple
            Residue IDs to extract. Each should be a (hetflag, resseq, icode) tuple.
        output_file : str
            Filepath to write the selected residues.
        """

        if self.entity.level != "C":
            logger.warning("Entity must be a Chain object.")
        else:
            res_sel = set()
            for res in residues:
                if res in self.entity.child_dict:
                    res_sel.add(self.entity[res])
                else:
                    logger.warning("Residue %s does not exist." % str(res))

            if len(res_sel) > 0:
                save_to_file(self.entity, 
                             output_file,
                             ResidueSelector(res_sel))
            else:
                logger.warning("No valid residue to extract.")



from collections import defaultdict

# Open Babel
from openbabel.openbabel import OBMol, OBSmartsPattern
from openbabel.pybel import Molecule as PybelMol
# RDKit
from rdkit.Chem import Mol as RDMol

from luna.util import stringcase as case
from luna.util.exceptions import MoleculeObjectTypeError
from luna.wrappers.base import MolWrapper

import logging
logger = logging.getLogger()


class ChemicalFeature():
    """Define chemical features as for example pharmacophore properties.

    Parameters
    ----------
    name : str
        The chemical feature name.
    """

    def __init__(self, name):
        self.name = name

    def format_name(self, case_func="sentencecase"):
        """Convert chemical feature names to another string case.

        Parameters
        ----------
        name : str
            The name of a string case function from
            :py:mod:`luna.util.stringcase`.
        """

        func = getattr(case, case_func)
        return func(self.name)

    # Special methods
    def __repr__(self):
        return "<Feature=%s>" % self.name

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(self, other.__class__):
            return self.name == other.name
        return False

    def __ne__(self, other):
        """Overrides the default implementation"""
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash(self.name)


class OBMolChemicalFeature:
    """Mimic :class:`rdkit.Chem.rdMolChemicalFeatures.MolChemicalFeature`
    for Open Babel.

    Parameters
    ----------
    family : str
        The family name, which is the term used by RDKit for chemical features.
    atom_ids : list
        List of atom identifiers.
    """

    def __init__(self, family, atom_ids):
        self.family = family
        self.atom_ids = atom_ids

    def GetAtomIds(self):
        """Get the IDs of the atoms that participate in the feature."""
        return self.atom_ids

    def GetFamily(self):
        """Get the family to which the feature belongs (e.g., donor,
        acceptor, etc.)"""
        return self.family


class FeatureExtractor:
    """Perceive chemical features from molecules.

    Parameters
    ----------
    feature_factory : \
        :class:`~rdkit.Chem.rdMolChemicalFeatures.MolChemicalFeatureFactory`
            An RDKit feature factory.

    Examples
    --------

    First, let's read a molecule (glutamine).

    >>> from luna.wrappers.base import MolWrapper
    >>> mol = MolWrapper.from_smiles("N[C@@H](CCC(N)=O)C(O)=O").unwrap()

    Now, create a feature factory and instantiate a new
    `FeatureExtractor` object.

    >>> from luna.util.default_values import ATOM_PROP_FILE
    >>> from rdkit.Chem import ChemicalFeatures
    >>> from luna.mol.features import FeatureExtractor
    >>> feature_factory = ChemicalFeatures.BuildFeatureFactory(ATOM_PROP_FILE)
    >>> feature_extractor = FeatureExtractor(feature_factory)

    Finally, you can extract features by group or atom.

    >>> features = feature_extractor.get_features_by_atoms(mol)
    >>> features = feature_extractor.get_features_by_groups(mol)

    """

    def __init__(self, feature_factory):
        self.feature_factory = feature_factory

    def get_features_by_atoms(self, mol_obj, atm_map=None):
        """Perceive chemical features from the molecule ``mol_obj`` by atom.

        Parameters
        ----------
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, \
                :class:`rdkit.Chem.rdchem.Mol`, or \
                :class:`openbabel.pybel.Molecule`
            The molecule.
        atm_map : dict
            A dictionary to map an atom's index to a different value.

        Returns
        -------
        atm_features : dict of {int : list of `ChemicalFeature`}
            Chemical features by atoms that are represented by their index or
            by a value from ``atm_map``.

        """
        if isinstance(mol_obj, MolWrapper):
            mol_obj = mol_obj.unwrap()

        if isinstance(mol_obj, RDMol):
            perceived_features = \
                self.feature_factory.GetFeaturesForMol(mol_obj)
        elif isinstance(mol_obj, OBMol):
            perceived_features = self._get_features_from_obmol(mol_obj)
        elif isinstance(mol_obj, PybelMol):
            perceived_features = self._get_features_from_obmol(mol_obj.OBMol)
        else:
            error_msg = ("Objects of type '%s' are not currently accepted."
                         % mol_obj.__class__)
            logger.exception(error_msg)
            raise MoleculeObjectTypeError(error_msg)

        atm_features = defaultdict(set)
        for f in perceived_features:
            for atm_idx in f.GetAtomIds():
                tmp_atm_idx = atm_idx
                if atm_map is not None:
                    if atm_idx in atm_map:
                        tmp_atm_idx = atm_map[atm_idx]
                    else:
                        logger.warning("There is no corresponding mapping to "
                                       "the index '%d'. It will be ignored."
                                       % atm_idx)

                feature = ChemicalFeature(f.GetFamily())
                atm_features[tmp_atm_idx].add(feature)

        return atm_features

    def get_features_by_groups(self, mol_obj, atm_map=None):
        """Perceive chemical features from the molecule ``mol_obj``
        by atom groups.

        Parameters
        ----------
        mol_obj : :class:`~luna.wrappers.base.MolWrapper`, \
                :class:`rdkit.Chem.rdchem.Mol`, or \
                :class:`openbabel.pybel.Molecule`
            The molecule.
        atm_map : dict
            A dictionary to map an atom's index to a different value.

        Returns
        -------
        grp_features : dict of {str : dict}
            Chemical features by groups. Each dictionary value is
            defined as follows:

            * ``atm_ids`` (list): list of atoms represented by their \
                    index or by a value from ``atm_map`` ;
            * ``features`` (list of `ChemicalFeature`): list of chemical \
                    features.

        """
        if isinstance(mol_obj, MolWrapper):
            mol_obj = mol_obj.unwrap()

        if isinstance(mol_obj, RDMol):
            perceived_features = \
                self.feature_factory.GetFeaturesForMol(mol_obj)
        elif isinstance(mol_obj, OBMol):
            perceived_features = self._get_features_from_obmol(mol_obj)
        elif isinstance(mol_obj, PybelMol):
            perceived_features = self._get_features_from_obmol(mol_obj.OBMol)
        else:
            error_msg = ("Objects of type '%s' are not currently accepted."
                         % mol_obj.__class__)
            logger.exception(error_msg)
            raise MoleculeObjectTypeError(error_msg)

        grp_features = {}
        for f in perceived_features:
            atm_ids = sorted(list(f.GetAtomIds()))

            if atm_map is not None:
                tmp_atm_ids = []
                for atm_id in atm_ids:
                    if atm_id in atm_map:
                        tmp_atm_ids.append(atm_map[atm_id])
                    else:
                        logger.warning("There is no corresponding mapping to "
                                       "the index '%d'. It will be ignored."
                                       % atm_id)
                atm_ids = tmp_atm_ids

            key = ','.join([str(x) for x in atm_ids])
            if key in grp_features:
                grp_obj = grp_features[key]
            else:
                grp_obj = {"atm_ids": atm_ids, "features": []}

            features = set(grp_obj["features"])
            feature = ChemicalFeature(f.GetFamily())
            features.add(feature)
            grp_obj["features"] = list(features)
            grp_features[key] = grp_obj

        return grp_features

    def _get_features_from_obmol(self, ob_mol):
        grp_features = defaultdict(set)
        for key, smarts in self.feature_factory.GetFeatureDefs().items():
            grp_type = key.split(".")[0]

            ob_smart = OBSmartsPattern()
            ob_smart.Init(str(smarts))
            ob_smart.Match(ob_mol)

            matches = [x for x in ob_smart.GetMapList()]
            if matches:
                for match in matches:
                    cur_ids = set(match)
                    exists = False
                    remove_ids = []

                    for ids in grp_features[grp_type]:
                        ids = set(ids)
                        # If there is any other group of the same type that
                        # already contains the current atoms.
                        if cur_ids.issubset(ids):
                            exists = True
                            # It just need to find one bigger group.
                            break
                        # If the current group contains atoms from already
                        # added group atoms.
                        elif ids.issubset(cur_ids):
                            exists = True
                            # Find all smaller groups to remove them.
                            remove_ids.append(tuple(ids))

                    if exists is False:
                        grp_features[grp_type].add(tuple(cur_ids))
                    elif remove_ids:
                        for ids in remove_ids:
                            grp_features[grp_type].remove(ids)
                        grp_features[grp_type].add(tuple(cur_ids))

        return [OBMolChemicalFeature(family, atom_ids)
                for family in grp_features
                for atom_ids in grp_features[family]]



import logging
from pathlib import Path
from io import StringIO

from luna.pdb.parser.base import PDBParser
from luna.util.exceptions import PDBNotReadError


logger = logging.getLogger()


def load_from_file(pdb_file, pdb_id=None):
    """
    Load a PDB structure from file using the default PDBParser configuration.

    Parameters
    ----------
    pdb_file : str or Path
        Path to the PDB file.
    pdb_id : str or None
        The structure identifier. If None, the filename will be used as id.
 
    Returns
    -------
    Structure
        The parsed Biopython structure object.

    Raises
    ------
    PDBNotReadError
        If the file cannot be parsed as a valid PDB structure.
    """
    
    try:
        pdb_id = pdb_id or Path(pdb_file).stem

        parser = PDBParser(PERMISSIVE=True, 
                           QUIET=True, 
                           FIX_EMPTY_CHAINS=True,
                           FIX_ATOM_NAME_CONFLICT=True, 
                           FIX_OBABEL_FLAGS=False)
        return parser.get_structure(pdb_id, pdb_file)
        
    except Exception as e:
        logger.exception(f"Failed to parse file '{pdb_file}': {str(e)}")
        raise PDBNotReadError(f"File '{pdb_file}' could not be parsed.")


def load_from_string(pdb_id: str, pdb_string: str):
    """
    Load a PDB structure from a string using the default PDBParser configuration.

    Parameters
    ----------
    pdb_id : str
        The structure identifier.
    pdb_string : str
        The full PDB content as a string.

    Returns
    -------
    Structure
        The parsed Biopython structure/

    Raises
    ------
    PDBNotReadError
        If the PDB string could not be parsed.
    """
    try:
        parser = PDBParser(PERMISSIVE=True,
                           QUIET=True,
                           FIX_EMPTY_CHAINS=True,
                           FIX_ATOM_NAME_CONFLICT=True,
                           FIX_OBABEL_FLAGS=False)

        handle = StringIO(pdb_string)
        return parser.get_structure_from_pdb_block(pdb_id, handle.read())

    except Exception as e:
        logger.exception(f"Failed to parse PDB string for '{pdb_id}': {str(e)}")
        raise PDBNotReadError(f"PDB string for '{pdb_id}' could not be parsed.")



from luna.util.default_values import ENTRY_SEPARATOR
from luna.util.exceptions import MoleculeNotFoundError, ChainNotFoundError



ENTITY_LEVEL_NAME = {"A": "Atom",
                     "R": "Residue",
                     "C": "Chain",
                     "M": "Model",
                     "S": "Structure"}


def get_entity_from_entry(entity, entry, model_id=0):
    """
    Retrieve a chain or residue from a structure based on an ``entry`` object.

    Parameters
    ----------
    entity : :class:`Bio.PDB.Entity.Entity`
        The PDB object to recover the target entry from.
    entry : :class:`luna.entry.Entry`
        Entry object defining the chain ID or residue ID.
    model_id : int
        Model id to search for the entry.

    Returns
    -------
    :class:`Bio.PDB.Entity.Entity`
        The matching Biopython entity.

    Raises
    ------
    MoleculeNotFoundError
        If the entry's molecule was not found in ``entity``.
    ChainNotFoundError
        If the entry's chain was not found in ``entity``.
    """

    structure = entity.get_parent_by_level("S")
    model = structure[model_id]

    if entry.chain_id not in model.child_dict:
        raise ChainNotFoundError(
            f"Chain '{entry.chain_id}' for entry '{entry.to_string(ENTRY_SEPARATOR)}' "
            f"not found in structure '{structure.get_id()}'."
        )
        
    chain = model[entry.chain_id]

    if entry.comp_name is not None and entry.comp_num is not None:
        ligand_key = entry.get_biopython_key()

        if ligand_key not in chain.child_dict:
            raise MoleculeNotFoundError(
                f"Ligand '{entry.to_string(ENTRY_SEPARATOR)}' not found in PDB '{structure.get_id()}'."
            )
        return chain[ligand_key]

    return chain
